from optimus.engines.base.create import BaseCreate
from optimus.helpers.types import *
import cudf
import pandas as pd

from optimus.engines.base.create import BaseCreate
from optimus.engines.dask_cudf.dataframe import DaskCUDFDataFrame
from optimus.helpers.types import DataFrameType, InternalDataFrameType


class Create(BaseCreate):

    @property
    def _pd(self):
        return cudf

    def _df_from_dfd(self, dfd, n_partitions=1, *args, **kwargs) -> 'DataFrameType':
        if isinstance(dfd, (pd.DataFrame,)):
            dfd = cudf.from_pandas(dfd)
        return DaskCUDFDataFrame(dfd, n_partitions, *args, **kwargs, op=self.op)



from abc import abstractmethod
import re

import dask
import dask.dataframe as dd
import hiurlparser
import numpy as np
from dask.delayed import delayed
from dask_ml.preprocessing import MinMaxScaler, StandardScaler
from sklearn.preprocessing import MaxAbsScaler
from dask_ml.impute import SimpleImputer
from optimus.engines.base.distributed.functions import DistributedBaseFunctions
from optimus.helpers.core import one_tuple_to_val, val_to_list
from optimus.infer import is_list


class DaskBaseFunctions(DistributedBaseFunctions):
    _engine = dask

    @staticmethod
    @property
    def _functions(self):
        return dd

    @staticmethod
    def _new_series(series, *args, **kwargs):

        if isinstance(series, dd.Series):
            return series
        
        index = kwargs.pop("index", None)

        if index is not None and hasattr(index, "to_frame"):
            index = index.to_frame()
            index["temp"] = series
            series = index["temp"]
            series.name = None
            return series

        return dd.from_array(series, *args, **kwargs)

    @staticmethod
    def compute(*args, **kwargs):
        result = dask.compute(*(*(a for a in args), *(kwargs[k] for k in kwargs)))
        return one_tuple_to_val(result)

    @abstractmethod
    def from_dataframe(self, dfd):
        pass

    @staticmethod
    def to_dataframe(dfd):
        return dfd.compute()

    @staticmethod
    def df_concat(df_list):
        return dd.concat(df_list, axis=0)

    def new_df(self, *args, **kwargs):
        if len(args) < 4 and all([k not in kwargs for k in ['dsk', 'name', 'meta', 'divisions']]):
            return self.from_dataframe(self._partition_engine.DataFrame(*args, **kwargs))
        return dd.DataFrame(*args, **kwargs)

    @staticmethod
    def dask_to_compatible(dfd):
        return dfd

    def sort_df(self, dfd, cols, ascending):
        for c, a in list(zip(cols, ascending))[::-1]:
            dfd = dfd.sort_values(c)
            if not a:
                dfd = self.reverse_df(dfd)
        return dfd.reset_index(drop=True)

    def reverse_df(self, dfd):
        @delayed
        def reverse_pdf(pdf):
            return pdf[::-1]

        ds = dfd.to_delayed()
        ds = [reverse_pdf(d) for d in ds][::-1]
        return dd.from_delayed(ds)

    def to_float(self, series):
        if getattr(series, "map_partitions", False):
            return self.map_partitions(series, self._to_float)
        else:
            return self._to_float(series)

    def to_integer(self, series, default=0):
        if getattr(series, "map_partitions", False):
            return self.map_partitions(series, self._to_integer, default=default)
        else:
            return self._to_integer(series, default=default)

    def to_datetime(self, series):
        if getattr(series, "map_partitions", False):
            return self.map_partitions(series, self._to_datetime)
        else:
            return self._to_float(series)

    def duplicated(self, dfd, keep, subset):
        return self.from_dataframe(self.to_dataframe(dfd).duplicated(keep=keep, subset=subset))

    def impute(self, series, strategy, fill_value):

        imputer = SimpleImputer(strategy=strategy, fill_value=fill_value)
        series_fit = series.dropna()
        if str(series.dtype) in self.constants.OBJECT_TYPES:
            series_fit = series_fit.astype(str)
        values = series_fit.values.reshape(-1, 1)
        if len(values):
            imputer.fit(values)
            return imputer.transform(series.fillna(np.nan).values.reshape(-1, 1))
        else:
            logger.warn("list to fit imputer is empty, try cols.fill_na instead.")
            return series

    @staticmethod
    def delayed(func):
        def wrapper(*args, **kwargs):
            return dask.delayed(func)(*args, **kwargs)

        return wrapper

    def from_delayed(self, delayed):
        return dd.from_delayed(delayed)

    def to_delayed(self, value):
        return value.to_delayed()

    def mode(self, series):
        # Uses delayed wrapper to avoid unexpected Dask errors
        # TODO should this be done in every static function called from Dask instead?
        @self.delayed
        def compute_mode(series):
            # dfd = series.rename(0).value_counts().reset_index(drop=False)
            # _max = dfd[0].max(skipna=True)
            # return dfd[dfd[0] == _max]['index'].rename(series.name)
            return series.mode()

        return compute_mode(series)

    def count_zeros(self, series):
        return int((self.to_float(series).values == 0).sum())

    def standard_scaler(self, series):
        dfd = StandardScaler().fit_transform(self.to_float(series).to_frame())
        return dfd[dfd.columns[0]]

    def max_abs_scaler(self, series):
        return MaxAbsScaler().fit_transform(self.compute(self.to_float(series)).values.reshape(-1, 1))

    def min_max_scaler(self, series):
        dfd = MinMaxScaler().fit_transform(self.to_float(series).to_frame())
        return dfd[dfd.columns[0]]

    # @staticmethod
    # def heatmap(df, bins):
    #     counts, edges = da.histogramdd((df['x'], df['y'].values), bins=bins)
    #     return counts, edges[0], edges[1]

    def domain(self, series):

        return self.to_string(series).map(lambda v: hiurlparser.parse_url(v)["domain"], na_action=None,
                                          meta=(series.name, "str"))

    def top_domain(self, series):
        return self.to_string(series).map(lambda v: hiurlparser.parse_url(v)["top_domain"], na_action=None,
                                          meta=(series.name, "str"))

    def sub_domain(self, series):
        return self.to_string(series).map(lambda v: hiurlparser.parse_url(v)["sub_domain"], na_action=None,
                                          meta=(series.name, "str"))

    def url_scheme(self, series):
        return self.to_string(series).map(lambda v: hiurlparser.parse_url(v)["protocol"], na_action=None,
                                          meta=(series.name, "str"))

    def url_path(self, series):
        return self.to_string(series).map(lambda v: hiurlparser.parse_url(v)["path"], na_action=None,
                                          meta=(series.name, "str"))

    def url_file(self, series):
        return self.to_string(series).map(lambda v: hiurlparser.parse_url(v)["file"], na_action=None,
                                          meta=(series.name, "str"))

    def url_query(self, series):
        return self.to_string(series).map(lambda v: hiurlparser.parse_url(v)["query"], na_action=None,
                                          meta=(series.name, "str"))

    def url_fragment(self, series):
        return self.to_string(series).map(lambda v: hiurlparser.parse_url(v)["fragment"], na_action=None,
                                          meta=(series.name, "str"))

    def host(self, series):
        return self.to_string(series).map(lambda v: hiurlparser.parse_url(v)["host"], na_action=None,
                                          meta=(series.name, "str"))

    def port(self, series):
        return self.to_string(series).map(lambda v: hiurlparser.parse_url(v)["port"], na_action=None,
                                          meta=(series.name, "str"))

    def _replace_string(self, series, to_replace, value, regex):
        
        if is_list(to_replace) and not is_list(value):
            if not regex:
                to_replace = [re.escape(x) for x in to_replace]
            to_replace = "|".join(to_replace)
            regex = True
        
        elif is_list(to_replace) and is_list(value):
            for _to_replace, _value in zip(to_replace, value):
                series = self._replace_string(series, _to_replace, _value, regex)
            return series

        return series.replace(to_replace, value, regex=regex)



from optimus.helpers.core import val_to_list
import sys
sys.path.append('../optimus')


alias_map_list = [

    # engine methods

    [["optimus", "engine", "createEngine"], "Optimus"],

    # op methods

    [["createDataframe", "dataframe"], "create.dataframe"],

    # connection methods

    ["connectMYSQL", "connect.mysql"],
    ["connectPostgres", "connect.postgres"],
    ["connectMSSQL", "connect.mssql"],
    ["connectRedshift", "connect.redshift"],
    ["connectSQLite", "connect.sqlite"],
    ["connectBigQuery", "connect.bigquery"],
    ["connectPresto", "connect.presto"],
    ["connectCassandra", "connect.cassandra"],
    ["connectRedis", "connect.redis"],
    ["connectOracle", "connect.oracle"],
    ["connectS3", "connect.s3"],
    ["connectLocal", "connect.local"],
    ["connectHDFS", "connect.hdfs"],
    ["connectGCS", "connect.gcs"],
    ["connectMAS", "connect.mas"],

    # string clustering methods

    ["setSuggestion", "set_suggestion"],
    ["setSuggestions", "set_suggestions"],

    # save methods

    ["saveFile", "save.file"],
    ["saveCSV", "save.csv"],
    ["saveXML", "save.xml"],
    ["saveJSON", "save.json"],
    ["saveExcel", "save.excel"],
    ["saveAvro", "save.avro"],
    ["saveParquet", "save.parquet"],
    ["saveORC", "save.orc"],
    ["saveHDF5", "save.hdf5"],

    # load methods

    ["loadFile", "load.file"],
    ["loadCSV", {"operation": "load.file", "file_type": "csv"}],
    ["loadXML", {"operation": "load.file", "file_type": "xml"}],
    ["loadJSON", {"operation": "load.file", "file_type": "json"}],
    ["loadExcel", {"operation": "load.file", "file_type": "excel"}],
    ["loadAvro", {"operation": "load.file", "file_type": "avro"}],
    ["loadParquet", {"operation": "load.file", "file_type": "parquet"}],
    ["loadORC", {"operation": "load.file", "file_type": "orc"}],
    ["loadZIP", {"operation": "load.file", "file_type": "zip"}],
    ["loadHDF5", {"operation": "load.file", "file_type": "hdf5"}],

    # methods

    ["profile", "profile"],
    [["profileColumns"], "profile.columns"],
    [["profileStats"], "profile.stats"],
    [["display", "toDict", "toObject"], "to_dict"],
    ["toJSON", "to_json"],
    ["size", "cols.size"],
    ["optimize", "cols.optimize"],
    ["run", "cols.run"],
    ["query", "cols.query"],
    ["partitions", "cols.partitions"],
    ["repartition", "cols.repartition"],
    [["print"], "ascii"],
    ["reset", "cols.reset"],
    [["agg", "aggregate", "groupyby", "groupBy"], "agg"],
    ["report", "cols.report"],
    [["selectColumns", "select"], "cols.select"],
    [["copyColumns", "copy"], "cols.copy"],
    [["dropColumns", "drop"], "cols.drop"],
    [["keepColumns", "keep"], "cols.keep"],
    [["renameColumns"], "cols.rename"],
    [["names", "columnNames", "columns"], "cols.names"],
    [["datatype", "dtype", "data_type", "datatypes", "dtypes", "data_types"], "cols.data_type"],
    [["patterCounts", "pattern_counts"], "cols.pattern_counts"],
    ["mad", "cols.mad"],
    ["min", "cols.min"],
    ["max", "cols.max"],
    ["mode", "cols.mode"],
    ["range", "cols.range"],
    ["percentile", "cols.percentile"],
    ["median", "cols.median"],
    ["kurtosis", "cols.kurtosis"],
    ["skew", "cols.skew"],
    ["mean", "cols.mean"],
    ["sum", "cols.sum"],
    ["cumsum", "cols.cumsum"],
    ["cumprod", "cols.cumprod"],
    ["cummax", "cols.cummax"],
    ["cummin", "cols.cummin"],
    ["var", "cols.var"],
    ["std", "cols.std"],
    ["iqr", "cols.iqr"],
    [["countNA", "count_na"], "cols.count_na"],
    ["unique", "cols.unique"],
    [["countUniques", "count_uniques"], "cols.count_uniques"],
    ["heatmap", "cols.heatmap"],
    ["hist", "cols.hist"],
    [["countMismatch", "count_mismatch"], "cols.count_mismatch"],
    ["quality", "cols.quality"],
    [
        ["inferProfilerDtypes", "infer_data_types"],
        "cols.infer_data_types",
    ],
    ["frequency", "cols.frequency"],
    ["boxplot", "cols.boxplot"],
    [["countZeros", "count_zeros"], "cols.count_zeros"],
    [["countRows", "length", "len"], "rows.count"],
    [["approxCount", "approx_count"], "rows.approx_count"],
    ["sample", "sample"],
    [["stratifiedSample", "stratified_sample"], "stratified_sample"],
    [["wordCount", "word_count", "cols.word_count"], "cols.word_count"],
    [["toTimestamp", "to_timestamp", "cols.to_timestamp"], "cols.to_timestamp"],
    [["setColumn", "set"], "cols.set"],
    [["castColumn", "cast"], "cols.cast"],
    [
        ["setDatatype", "setDtype", "set_data_type", "cols.set_data_type"],
        "cols.set_data_type",
    ],
    [
        ["unsetDatatype", "unsetDtype", "unset_data_type", "cols.unset_data_type"],
        "cols.unset_data_type",
    ],
    ["replace", "cols.replace"],
    [["replaceRegex", "replace_regex"], "cols.replace_regex"],
    ["pattern", "cols.pattern"],
    ["join", "cols.join"],
    ["move", "cols.move"],
    ["sort", "cols.sort"],
    ["item", "cols.item"],
    ["get", "cols.get"],
    ["abs", "cols.abs"],
    ["exp", "cols.exp"],
    ["mod", "cols.mod"],
    ["log", "cols.log"],
    ["ln", "cols.ln"],
    ["pow", "cols.pow"],
    ["sqrt", "cols.sqrt"],
    ["reciprocal", "cols.reciprocal"],
    ["round", "cols.round"],
    ["floor", "cols.floor"],
    ["ceil", "cols.ceil"],
    ["sin", "cols.sin"],
    ["cos", "cols.cos"],
    ["tan", "cols.tan"],
    ["asin", "cols.asin"],
    ["acos", "cols.acos"],
    ["atan", "cols.atan"],
    ["sinh", "cols.sinh"],
    ["cosh", "cols.cosh"],
    ["tanh", "cols.tanh"],
    ["asinh", "cols.asinh"],
    ["acosh", "cols.acosh"],
    ["atanh", "cols.atanh"],
    ["extract", "cols.extract"],
    ["slice", "cols.slice"],
    ["left", "cols.left"],
    ["right", "cols.right"],
    ["mid", "cols.mid"],
    [["toFloat", "to_float"], "cols.to_float"],
    [["toInteger", "to_integer"], "cols.to_integer"],
    [["toBoolean", "to_boolean"], "cols.to_boolean"],
    [["toString", "to_string"], "cols.to_string"],
    ["matchRegex", "cols.match"],
    [["lowercase", "lower"], "cols.lower"],
    [["uppercase", "upper"], "cols.upper"],
    [["propercase", "proper"], "cols.proper"],
    ["title", "cols.title"],
    ["capitalize", "cols.capitalize"],
    ["pad", "cols.pad"],
    ["trim", "cols.trim"],
    [["stripHtml", "strip_html"], "cols.strip_html"],
    [["dateFormat", "date_format"], "cols.date_format"],
    [["formatDate", "format_date"], "cols.format_date"],
    [["wordTokenize", "word_tokenize"], "cols.word_tokenize"],
    [["wordCount", "word_count"], "cols.word_count"],
    ["len", "cols.len"],
    ["reverseString", "cols.reverse"],
    ["remove", "cols.remove"],
    [
        ["normalizeCharacters", "normalizeChars", "normalize_chars"],
        "cols.normalize_chars",
    ],
    [["removeNumbers", "remove_numbers"], "cols.remove_numbers"],
    [["removeWhiteSpaces", "remove_white_spaces"], "cols.remove_white_spaces"],
    [["removeStopWords", "remove_stopwords"], "cols.remove_stopwords"],
    [["removeUrls", "remove_urls"], "cols.remove_urls"],
    [["normalizeSpaces", "normalize_spaces"], "cols.normalize_spaces"],
    [["removeSpecialChars", "remove_special_chars"], "cols.remove_special_chars"],
    [["toDatetime", "to_datetime"], "cols.to_datetime"],
    ["year", "cols.year"],
    ["month", "cols.month"],
    ["day", "cols.day"],
    ["hour", "cols.hour"],
    ["minute", "cols.minute"],
    ["second", "cols.second"],
    ["weekday", "cols.weekday"],
    [["yearsBetween", "years_between"], "cols.years_between"],
    [["lemmatizeVerb", "lemmatize_verb"], "cols.lemmatize_verb"],
    ["impute", "cols.impute"],
    [["fillNa", "fill_na"], "cols.fill_na"],
    ["countCols", "cols.count"],
    ["add", "cols.add"],
    ["sub", "cols.sub"],
    ["mul", "cols.mul"],
    ["div", "cols.div"],
    [["zScore", "z_score"], "cols.z_score"],
    [["modifiedZScore", "modified_z_score"], "cols.modified_z_score"],
    [["minMaxScaler", "min_max_scaler"], "cols.min_max_scaler"],
    [["standardScaler", "standard_scaler"], "cols.standard_scaler"],
    [["maxAbsScaler", "max_abs_scaler"], "cols.max_abs_scaler"],
    ["nest", "cols.nest"],
    ["unnest", "cols.unnest"],
    ["qcut", "cols.qcut"],
    ["cut", "cols.cut"],
    ["clip", "cols.clip"],
    [["stringToIndex", "string_to_index"], "cols.string_to_index"],
    [["indexToString", "index_to_string"], "cols.index_to_string"],
    [["topDomain", "top_domain"], "cols.top_domain"],
    ["domain", "cols.domain"],
    [["URLScheme", "url_scheme"], "cols.url_scheme"],
    [["URLParams", "url_params"], "cols.url_params"],
    [["URLPath", "url_path"], "cols.url_path"],
    ["port", "cols.port"],
    [["URLQuery", "url_query"], "cols.url_query"],
    ["subdomain", "cols.subdomain"],
    [["emailUsername", "email_username"], "cols.email_username"],
    [["emailDomain", "email_domain"], "cols.email_domain"],
    ["missing", "cols.missing"],
    ["fingerprint", "cols.fingerprint"],
    [["nGramFingerprint", "n_gram_fingerprint"], "cols.n_gram_fingerprint"],
    ["metaphone", "cols.metaphone"],
    ["nysiis", "cols.nysiis"],
    [["matchRatingCodex", "match_rating_codex"], "cols.match_rating_codex"],
    [["doubleMethaphone", "double_metaphone"], "cols.double_metaphone"],
    ["soundex", "cols.soundex"],
    ["append", "rows.append"],
    [["greaterThan", "greater_than"], "rows.greater_than"],
    [["greaterThanEqual", "greater_than_equal"], "rows.greater_than_equal"],
    [["lessThan", "less_than"], "rows.less_than"],
    [["lessThanEqual", "less_than_equal"], "rows.less_than_equal"],
    ["equal", "rows.equal"],
    [["notEqual", "not_equal"], "rows.not_equal"],
    ["missing", "rows.missing"],
    ["mismatch", "rows.mismatch"],
    ["match", "rows.match"],
    ["find", "rows.find"],
    ["selectRows", "rows.select"],
    ["sortRows", "rows.sort"],
    ["reverse", "rows.reverse"],
    ["dropRows", "rows.drop"],
    ["between", "rows.between"],
    [["dropNA", "drop_na"], "rows.drop_missings"],
    [["dropDuplicated", "drop_duplicated"], "rows.drop_duplicated"],
    ["limit", "rows.limit"],
    [["isIn", "is_in"], "rows.is_in"]
]

alias_map = {}

for alias_group, value in alias_map_list:
    for alias in val_to_list(alias_group):
        alias_map.update({alias: value})


def use_alias(body: dict):

    if body["operation"] in alias_map:
        found = alias_map[body["operation"]]
        if isinstance(found, str):
            body["operation"] = found
        else:
            body.update(found)

    return body


def default_to_engine(body: dict):

    if (not "operation" in body) and ("engine" in body):
        body["operation"] = "Optimus"

    return body


def default_to_display(body: dict):

    if (not "operation" in body) and ("source" in body):
        body["operation"] = "to_dict"

    return body


preparers = [
    default_to_engine,
    default_to_display,
    use_alias
]


def prepare(body: dict):
    for preparer in preparers:
        body = preparer(body)

    return body



from copy import deepcopy as deep_copy

from glom import glom, assign, delete

from optimus.helpers.core import val_to_list
from optimus.infer import is_dict, is_list_value

ACTIONS_PATH = "transformations.actions"

import orjson


def deepcopy(data):
    """
    Deep copy of a dictionary or list. Standard python deepcopy is 10x slower.
    :param data:
    :return:
    """
    try:
        return orjson.loads(orjson.dumps(data))
    except:
        return deep_copy(data)


class Meta:

    @staticmethod
    def set(meta, spec=None, value=None, missing=dict) -> dict:
        """
        Set metadata in a dataframe columns
        :param meta: Meta data to be modified
        :param spec: path to the key to be modified
        :param value: dict value
        :param missing:
        :return:
        """
        if spec is not None:
            data = deepcopy(meta)
            assign(data, spec, value, missing=missing)
        else:
            data = value

        return data

    @staticmethod
    def reset(meta, spec=None) -> dict:
        """
        Set metadata in a dataframe columns
        :param meta: Meta data to be modified
        :param spec: path to the key to be modified
        :return:
        """
        if spec is not None:
            data = deepcopy(meta)
            delete(data, spec, ignore_missing=True)
        else:
            data = meta

        return data

    @staticmethod
    def get(meta, spec=None) -> dict:
        """
        Get metadata from a dataframe column
        :param meta:Meta data to be modified
        :param spec: path to the key to be modified
        :return: dict
        """
        if spec is not None:
            data = glom(meta, spec, skip_exc=KeyError)
        else:
            data = meta
        return deepcopy(data) if isinstance(data, dict) else data

    @staticmethod
    def reset_actions(meta, cols="*"):
        """
        Reset the data frame metadata
        :param meta: Meta data to be modified
        :return:
        """

        if cols == "*":
            return Meta.set(meta, ACTIONS_PATH, [])
        else:
            actions = Meta.get(meta, ACTIONS_PATH) or []
            actions = [action for action in actions if action["columns"] not in (cols or [])]
            return Meta.set(meta, ACTIONS_PATH, actions)

    @staticmethod
    def columns(meta, value) -> dict:
        """
        Shortcut to cache the columns in a dataframe
        :param meta: Meta data to be modified
        :param value:
        :return: dict (Meta)
        """
        value = val_to_list(value)
        for v in value:
            meta = Meta.update(meta, "transformations.columns", v, list)
        return meta

    @staticmethod
    def action(meta, name, value) -> dict:
        """
        Shortcut to add actions to a dataframe
        :param meta: Meta data to be modified
        :param name: Action name
        :param value: Value to be added
        :return: dict (Meta)
        """
        if not is_list_value(value):
            value = [value]
        for v in value:
            meta = Meta.update(meta, ACTIONS_PATH, {"name": name, "columns": v}, list)
        return meta

    @staticmethod
    def update(meta, path, value, default=list) -> dict:
        """
        Update meta data in a key
        :param meta: Meta data to be modified
        :param path: Path inside the dict to be modified
        :param value: New key value
        :param default:
        :return: dict (Meta)
        """

        new_meta = deepcopy(meta)

        if new_meta is None:
            new_meta = {}

        elements = path.split(".")
        _element = new_meta
        for i, ele in enumerate(elements):
            if ele not in _element and not len(elements) - 1 == i:
                _element[ele] = {}

            if len(elements) - 1 == i:
                if default is list:
                    _element[ele] = _element.get(ele, [])
                    _element[ele].append(value)
                elif default is dict:
                    _element[ele] = _element.get(ele, {})
                    _element[ele].update(value)
            else:
                _element = _element[ele]

        return new_meta

    @staticmethod
    def drop_columns(meta, cols):
        for col in cols:
            meta = Meta.reset(meta, f'columns_data_types.{col}')
            meta = Meta.reset(meta, f'max_cell_length.{col}')
            meta = Meta.reset(meta, f'profile.columns.{col}')
        return meta

    @staticmethod
    def select_columns(meta, cols):
        all_cols = []

        for _cols in ['columns_data_types', 'max_cell_length', 'profile.columns']:
            found = Meta.get(meta, _cols)
            if is_dict(found):
                all_cols += list(found.keys())

        to_drop = list(set(all_cols) - set(cols))
        return Meta.drop_columns(meta, to_drop)



import numpy as np

from optimus.infer import is_dict, is_dict_of_one_element, is_list_value, is_list_of_one_element


def tuple_to_dict(value):
    """
    Convert tuple to dict
    :param value: tuple to be converted
    :return:
    """

    return dict((x, y) for x, y in value)


def format_dict(_dict, tidy=True):
    """
    This function format a dict. If the main dict or a deep dict has only on element
     {"col_name":{0.5: 200}} we get 200
    :param _dict: dict to be formatted
    :param tidy:
    :return:
    """

    if tidy is True:
        levels = 2
        while (levels>=0):
            levels -= 1
            if is_list_of_one_element(_dict):
                _dict = _dict[0]
            elif is_dict_of_one_element(_dict):
                _dict = next(iter(_dict.values()))
            else:
                return _dict
                
    else:
        if is_list_of_one_element(_dict):
            return _dict[0]
        else:
            return _dict

def convert_numpy(value):
    if isinstance(value, (dict,)):
        for key in value:
            value[key] = convert_numpy(value[key])
        return value
    elif isinstance(value, (list, set, tuple)):
        return value.__class__(map(convert_numpy, value))
    elif isinstance(value, (np.generic,)):
        return np.asscalar(value)
    elif hasattr(value, "to_pydatetime"):
        return value.to_pydatetime()
    else:
        return value

# def cudf_series_to_pandas(serie):
#     return serie.to_pandas()

def dask_dataframe_to_dask_cudf(df):
    import cudf
    return df.map_partitions(cudf.DataFrame.from_pandas)

# To cudf
def dask_dataframe_to_cudf(df):
    return pandas_to_cudf(dask_dataframe_to_pandas(df))

# def dask_cudf_to_cudf(df):
#     return df.compute()


# To Pandas
def spark_to_pandas(df):
    return df.toPandas()


def dask_cudf_to_pandas(df):
    return df.map_partitions(lambda df: df.to_pandas())


def dask_dataframe_to_pandas(df):
    return df.compute()


def cudf_to_pandas(df):
    return df.to_pandas()

def cudf_to_dask_cudf(df, n_partitions=1):
    import dask_cudf
    return dask_cudf.from_cudf(df, npartitions=1)

# def cudf_to_cupy_arr(df):
#     import cupy as cp
#     return cp.fromDlpack(df.to_dlpack())

def pandas_to_cudf(df):
    import cudf
    return cudf.from_pandas(df)


def pandas_to_dask_dataframe(pdf, n_partitions=1):
    from dask import dataframe as dd
    return dd.from_pandas(pdf, npartitions=n_partitions)


def pandas_to_vaex_dataframe(pdf, n_partitions=1):
    import vaex
    return vaex.from_pandas(pdf)


def pandas_to_dask_cudf_dataframe(pdf, n_partitions=1):
    import cudf
    import dask_cudf
    # Seems that from_cudf also accepts pandas
    cdf = cudf.DataFrame.from_pandas(pdf)
    return dask_cudf.from_cudf(cdf, npartitions=n_partitions)

def pandas_to_spark_dataframe(pdf, n_partitions=1):
    import databricks.koalas as ks
    return ks.from_pandas(pdf)



from enum import Enum

from optimus.helpers.functions import singleton

from optimus.engines.base.io.drivers.abstract_driver import AbstractDriver
from optimus.engines.base.io.properties import DriverProperties


@singleton
class OracleDriver(AbstractDriver):
    """Oracle Database"""

    def properties(self) -> Enum:
        return DriverProperties.ORACLE

    def url(self, *args, **kwargs) -> str:
        if kwargs["oracle_sid"] is not None:
            return f"""jdbc:oracle+cx_oracle:thin:@{kwargs["host"]}:{kwargs["port"]}/{kwargs["oracle_sid"]}"""
        elif kwargs["oracle_service_name"] is not None:
            return f"""jdbc:oracle+cx_oracle:thin:@//{kwargs["host"]}:{kwargs["port"]}/{kwargs[
                "oracle_service_name"]}"""
        else:
            return f"""jdbc:oracle+cx_oracle:thin:@//{kwargs["oracle_tns"]}"""

    def table_names_query(self, *args, **kwargs) -> str:
        return """
            SELECT table_name, extractvalue(xmltype( dbms_xmlgen.getxml('select count(*) c from '||table_name)), '/ROWSET/ROW/C') count
            FROM user_tables ORDER BY table_name
        """

    def table_name_query(self, *args, **kwargs) -> str:
        return "SELECT table_name as 'table_name' FROM user_tables"

    def count_query(self, *args, **kwargs) -> str:
        return "SELECT COUNT(*) COUNT FROM " + kwargs["db_table"]

    def primary_key_query(self, *args, **kwargs) -> str:
        pass

    def min_max_query(self, *args, **kwargs) -> str:
        return f"""SELECT min({kwargs["partition_column"]}) AS min, max({kwargs["partition_column"]}) AS max FROM {
        kwargs["table_name"]} """



import datetime
import numpy as np
from optimus.tests.base import TestBase
from optimus.helpers.json import json_encoding
from optimus.helpers.functions import deep_sort, df_dicts_equal, results_equal


def Timestamp(t):
    return datetime.datetime.strptime(t, "%Y-%m-%d %H:%M:%S")


NaT = np.datetime64('NaT')
nan = float("nan")
inf = float("inf")


class TestStatisticsPandas(TestBase):
    config = {'engine': 'pandas'}
    dict = {('NullType', 'object'): [None, None, None, None, None, None], ('attributes', 'object'): [[8.5344, 4300.0], [5.334, 2000.0], [7.9248, 4000.0], [3.9624, 1800.0], [None, 5700.0], [91.44, None]], ('date arrival', 'object'): ['1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10'], ('function(binary)', 'object'): [bytearray(b'Leader'), bytearray(b'Espionage'), bytearray(b'Security'), bytearray(b'First Lieutenant'), bytearray(b'None'), bytearray(b'Battle Station')], ('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 13.0, nan, 300.0], ('japanese name', 'object'): [['Inochi', 'Convoy'], ['Bumble', 'Goldback'], ['Roadbuster'], ['Meister'], ['Megatron'], ['Metroflex']], ('last date seen', 'object'): ['2016/09/10', '2015/08/10', '2014/07/10', '2013/06/10', '2012/05/10', '2011/04/10'], ('last position seen', 'object'): ['19.442735,-99.201111', '10.642707,-71.612534', '37.789563,-122.400356', '33.670666,-117.841553', None, None], ('rank', 'int64'): [10, 7, 7, 8, 10, 8], ('Cybertronian', 'bool'): [True, True, True, True, True, False], ('Date Type', 'datetime64[ns]'): [Timestamp('2016-09-10 00:00:00'), Timestamp('2015-08-10 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2013-06-24 00:00:00'), Timestamp('2012-05-10 00:00:00'), Timestamp('2011-04-10 00:00:00')], ('age', 'int64'): [5000000, 5000000, 5000000, 5000000, 5000000, 5000000], ('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$'], ('timestamp', 'datetime64[ns]'): [Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00')], ('weight(t)', 'float64'): [4.3, 2.0, 4.0, 1.8, 5.7, nan]}
    maxDiff = None

    def test_cols_cummax_all(self):
        df = self.df.copy()
        result = df.cols.cummax(cols='*')
        expected = self.create_dataframe(data={('NullType', 'float64'): [nan, nan, nan, nan, nan, nan], ('attributes', 'float64'): [nan, nan, nan, nan, nan, nan], ('date arrival', 'float64'): [nan, nan, nan, nan, nan, nan], ('function(binary)', 'float64'): [nan, nan, nan, nan, nan, nan], ('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 26.0, nan, 300.0], ('japanese name', 'float64'): [nan, nan, nan, nan, nan, nan], ('last date seen', 'float64'): [nan, nan, nan, nan, nan, nan], ('last position seen', 'float64'): [nan, nan, nan, nan, nan, nan], ('rank', 'float64'): [10.0, 10.0, 10.0, 10.0, 10.0, 10.0], ('Cybertronian', 'float64'): [1.0, 1.0, 1.0, 1.0, 1.0, 1.0], ('Date Type', 'float64'): [1.4734656e+18, 1.4734656e+18, 1.4734656e+18, 1.4734656e+18, 1.4734656e+18, 1.4734656e+18], ('age', 'float64'): [5000000.0, 5000000.0, 5000000.0, 5000000.0, 5000000.0, 5000000.0], ('function', 'float64'): [nan, nan, nan, nan, nan, nan], ('names', 'float64'): [nan, nan, nan, nan, nan, nan], ('timestamp', 'float64'): [1.403568e+18, 1.403568e+18, 1.403568e+18, 1.403568e+18, 1.403568e+18, 1.403568e+18], ('weight(t)', 'float64'): [4.3, 4.3, 4.3, 4.3, 5.7, nan]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_cummax_multiple(self):
        df = self.df.copy()
        result = df.cols.cummax(cols=['height(ft)', 'age', 'rank'])
        expected = self.create_dataframe(data={('NullType', 'object'): [None, None, None, None, None, None], ('attributes', 'object'): [[8.5344, 4300.0], [5.334, 2000.0], [7.9248, 4000.0], [3.9624, 1800.0], [None, 5700.0], [91.44, None]], ('date arrival', 'object'): ['1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10'], ('function(binary)', 'object'): [bytearray(b'Leader'), bytearray(b'Espionage'), bytearray(b'Security'), bytearray(b'First Lieutenant'), bytearray(b'None'), bytearray(b'Battle Station')], ('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 26.0, nan, 300.0], ('japanese name', 'object'): [['Inochi', 'Convoy'], ['Bumble', 'Goldback'], ['Roadbuster'], ['Meister'], ['Megatron'], ['Metroflex']], ('last date seen', 'object'): ['2016/09/10', '2015/08/10', '2014/07/10', '2013/06/10', '2012/05/10', '2011/04/10'], ('last position seen', 'object'): ['19.442735,-99.201111', '10.642707,-71.612534', '37.789563,-122.400356', '33.670666,-117.841553', None, None], ('rank', 'float64'): [10.0, 10.0, 10.0, 10.0, 10.0, 10.0], ('Cybertronian', 'bool'): [True, True, True, True, True, False], ('Date Type', 'datetime64[ns]'): [Timestamp('2016-09-10 00:00:00'), Timestamp('2015-08-10 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2013-06-24 00:00:00'), Timestamp('2012-05-10 00:00:00'), Timestamp('2011-04-10 00:00:00')], ('age', 'float64'): [5000000.0, 5000000.0, 5000000.0, 5000000.0, 5000000.0, 5000000.0], ('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$'], ('timestamp', 'datetime64[ns]'): [Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00')], ('weight(t)', 'float64'): [4.3, 2.0, 4.0, 1.8, 5.7, nan]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_cummax_numeric(self):
        df = self.df.copy()
        result = df.cols.cummax(cols='weight(t)')
        expected = self.create_dataframe(data={('NullType', 'object'): [None, None, None, None, None, None], ('attributes', 'object'): [[8.5344, 4300.0], [5.334, 2000.0], [7.9248, 4000.0], [3.9624, 1800.0], [None, 5700.0], [91.44, None]], ('date arrival', 'object'): ['1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10'], ('function(binary)', 'object'): [bytearray(b'Leader'), bytearray(b'Espionage'), bytearray(b'Security'), bytearray(b'First Lieutenant'), bytearray(b'None'), bytearray(b'Battle Station')], ('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 13.0, nan, 300.0], ('japanese name', 'object'): [['Inochi', 'Convoy'], ['Bumble', 'Goldback'], ['Roadbuster'], ['Meister'], ['Megatron'], ['Metroflex']], ('last date seen', 'object'): ['2016/09/10', '2015/08/10', '2014/07/10', '2013/06/10', '2012/05/10', '2011/04/10'], ('last position seen', 'object'): ['19.442735,-99.201111', '10.642707,-71.612534', '37.789563,-122.400356', '33.670666,-117.841553', None, None], ('rank', 'int64'): [10, 7, 7, 8, 10, 8], ('Cybertronian', 'bool'): [True, True, True, True, True, False], ('Date Type', 'datetime64[ns]'): [Timestamp('2016-09-10 00:00:00'), Timestamp('2015-08-10 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2013-06-24 00:00:00'), Timestamp('2012-05-10 00:00:00'), Timestamp('2011-04-10 00:00:00')], ('age', 'int64'): [5000000, 5000000, 5000000, 5000000, 5000000, 5000000], ('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$'], ('timestamp', 'datetime64[ns]'): [Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00')], ('weight(t)', 'float64'): [4.3, 4.3, 4.3, 4.3, 5.7, nan]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_cummin_all(self):
        df = self.df.copy()
        result = df.cols.cummin(cols='*')
        expected = self.create_dataframe(data={('NullType', 'float64'): [nan, nan, nan, nan, nan, nan], ('attributes', 'float64'): [nan, nan, nan, nan, nan, nan], ('date arrival', 'float64'): [nan, nan, nan, nan, nan, nan], ('function(binary)', 'float64'): [nan, nan, nan, nan, nan, nan], ('height(ft)', 'float64'): [-28.0, -28.0, -28.0, -28.0, nan, -28.0], ('japanese name', 'float64'): [nan, nan, nan, nan, nan, nan], ('last date seen', 'float64'): [nan, nan, nan, nan, nan, nan], ('last position seen', 'float64'): [nan, nan, nan, nan, nan, nan], ('rank', 'float64'): [10.0, 7.0, 7.0, 7.0, 7.0, 7.0], ('Cybertronian', 'float64'): [1.0, 1.0, 1.0, 1.0, 1.0, 0.0], ('Date Type', 'float64'): [1.4734656e+18, 1.4391648e+18, 1.403568e+18, 1.372032e+18, 1.336608e+18, 1.3023936e+18], ('age', 'float64'): [5000000.0, 5000000.0, 5000000.0, 5000000.0, 5000000.0, 5000000.0], ('function', 'float64'): [nan, nan, nan, nan, nan, nan], ('names', 'float64'): [nan, nan, nan, nan, nan, nan], ('timestamp', 'float64'): [1.403568e+18, 1.403568e+18, 1.403568e+18, 1.403568e+18, 1.403568e+18, 1.403568e+18], ('weight(t)', 'float64'): [4.3, 2.0, 2.0, 1.8, 1.8, nan]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_cummin_multiple(self):
        df = self.df.copy()
        result = df.cols.cummin(cols=['height(ft)', 'age', 'rank'])
        expected = self.create_dataframe(data={('NullType', 'object'): [None, None, None, None, None, None], ('attributes', 'object'): [[8.5344, 4300.0], [5.334, 2000.0], [7.9248, 4000.0], [3.9624, 1800.0], [None, 5700.0], [91.44, None]], ('date arrival', 'object'): ['1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10'], ('function(binary)', 'object'): [bytearray(b'Leader'), bytearray(b'Espionage'), bytearray(b'Security'), bytearray(b'First Lieutenant'), bytearray(b'None'), bytearray(b'Battle Station')], ('height(ft)', 'float64'): [-28.0, -28.0, -28.0, -28.0, nan, -28.0], ('japanese name', 'object'): [['Inochi', 'Convoy'], ['Bumble', 'Goldback'], ['Roadbuster'], ['Meister'], ['Megatron'], ['Metroflex']], ('last date seen', 'object'): ['2016/09/10', '2015/08/10', '2014/07/10', '2013/06/10', '2012/05/10', '2011/04/10'], ('last position seen', 'object'): ['19.442735,-99.201111', '10.642707,-71.612534', '37.789563,-122.400356', '33.670666,-117.841553', None, None], ('rank', 'float64'): [10.0, 7.0, 7.0, 7.0, 7.0, 7.0], ('Cybertronian', 'bool'): [True, True, True, True, True, False], ('Date Type', 'datetime64[ns]'): [Timestamp('2016-09-10 00:00:00'), Timestamp('2015-08-10 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2013-06-24 00:00:00'), Timestamp('2012-05-10 00:00:00'), Timestamp('2011-04-10 00:00:00')], ('age', 'float64'): [5000000.0, 5000000.0, 5000000.0, 5000000.0, 5000000.0, 5000000.0], ('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$'], ('timestamp', 'datetime64[ns]'): [Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00')], ('weight(t)', 'float64'): [4.3, 2.0, 4.0, 1.8, 5.7, nan]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_cummin_numeric(self):
        df = self.df.copy()
        result = df.cols.cummin(cols='weight(t)')
        expected = self.create_dataframe(data={('NullType', 'object'): [None, None, None, None, None, None], ('attributes', 'object'): [[8.5344, 4300.0], [5.334, 2000.0], [7.9248, 4000.0], [3.9624, 1800.0], [None, 5700.0], [91.44, None]], ('date arrival', 'object'): ['1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10'], ('function(binary)', 'object'): [bytearray(b'Leader'), bytearray(b'Espionage'), bytearray(b'Security'), bytearray(b'First Lieutenant'), bytearray(b'None'), bytearray(b'Battle Station')], ('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 13.0, nan, 300.0], ('japanese name', 'object'): [['Inochi', 'Convoy'], ['Bumble', 'Goldback'], ['Roadbuster'], ['Meister'], ['Megatron'], ['Metroflex']], ('last date seen', 'object'): ['2016/09/10', '2015/08/10', '2014/07/10', '2013/06/10', '2012/05/10', '2011/04/10'], ('last position seen', 'object'): ['19.442735,-99.201111', '10.642707,-71.612534', '37.789563,-122.400356', '33.670666,-117.841553', None, None], ('rank', 'int64'): [10, 7, 7, 8, 10, 8], ('Cybertronian', 'bool'): [True, True, True, True, True, False], ('Date Type', 'datetime64[ns]'): [Timestamp('2016-09-10 00:00:00'), Timestamp('2015-08-10 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2013-06-24 00:00:00'), Timestamp('2012-05-10 00:00:00'), Timestamp('2011-04-10 00:00:00')], ('age', 'int64'): [5000000, 5000000, 5000000, 5000000, 5000000, 5000000], ('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$'], ('timestamp', 'datetime64[ns]'): [Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00')], ('weight(t)', 'float64'): [4.3, 2.0, 2.0, 1.8, 1.8, nan]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_cumprod_all(self):
        df = self.df.copy()
        result = df.cols.cumprod(cols='*')
        expected = self.create_dataframe(data={('NullType', 'float64'): [nan, nan, nan, nan, nan, nan], ('attributes', 'float64'): [nan, nan, nan, nan, nan, nan], ('date arrival', 'float64'): [nan, nan, nan, nan, nan, nan], ('function(binary)', 'float64'): [nan, nan, nan, nan, nan, nan], ('height(ft)', 'float64'): [-28.0, -476.0, -12376.0, -160888.0, nan, -48266400.0], ('japanese name', 'float64'): [nan, nan, nan, nan, nan, nan], ('last date seen', 'float64'): [nan, nan, nan, nan, nan, nan], ('last position seen', 'float64'): [nan, nan, nan, nan, nan, nan], ('rank', 'float64'): [10.0, 70.0, 490.0, 3920.0, 39200.0, 313600.0], ('Cybertronian', 'float64'): [1.0, 1.0, 1.0, 1.0, 1.0, 0.0], ('Date Type', 'float64'): [1.4734656e+18, 2.12055982553088e+36, 2.9763499132007264e+54, 4.083647324108619e+72, 5.458235682582173e+90, 7.108771220286654e+108], ('age', 'float64'): [5000000.0, 25000000000000.0, 1.25e+20, 6.25e+26, 3.125e+33, 1.5625e+40], ('function', 'float64'): [nan, nan, nan, nan, nan, nan], ('names', 'float64'): [nan, nan, nan, nan, nan, nan], ('timestamp', 'float64'): [1.403568e+18, 1.970003130624e+36, 2.7650333540436665e+54, 3.880912334668361e+72, 5.447124363745802e+90, 7.645409448973968e+108], ('weight(t)', 'float64'): [4.3, 8.6, 34.4, 61.92, 352.944, nan]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_cumprod_multiple(self):
        df = self.df.copy()
        result = df.cols.cumprod(cols=['height(ft)', 'age', 'rank'])
        expected = self.create_dataframe(data={('NullType', 'object'): [None, None, None, None, None, None], ('attributes', 'object'): [[8.5344, 4300.0], [5.334, 2000.0], [7.9248, 4000.0], [3.9624, 1800.0], [None, 5700.0], [91.44, None]], ('date arrival', 'object'): ['1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10'], ('function(binary)', 'object'): [bytearray(b'Leader'), bytearray(b'Espionage'), bytearray(b'Security'), bytearray(b'First Lieutenant'), bytearray(b'None'), bytearray(b'Battle Station')], ('height(ft)', 'float64'): [-28.0, -476.0, -12376.0, -160888.0, nan, -48266400.0], ('japanese name', 'object'): [['Inochi', 'Convoy'], ['Bumble', 'Goldback'], ['Roadbuster'], ['Meister'], ['Megatron'], ['Metroflex']], ('last date seen', 'object'): ['2016/09/10', '2015/08/10', '2014/07/10', '2013/06/10', '2012/05/10', '2011/04/10'], ('last position seen', 'object'): ['19.442735,-99.201111', '10.642707,-71.612534', '37.789563,-122.400356', '33.670666,-117.841553', None, None], ('rank', 'float64'): [10.0, 70.0, 490.0, 3920.0, 39200.0, 313600.0], ('Cybertronian', 'bool'): [True, True, True, True, True, False], ('Date Type', 'datetime64[ns]'): [Timestamp('2016-09-10 00:00:00'), Timestamp('2015-08-10 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2013-06-24 00:00:00'), Timestamp('2012-05-10 00:00:00'), Timestamp('2011-04-10 00:00:00')], ('age', 'float64'): [5000000.0, 25000000000000.0, 1.25e+20, 6.25e+26, 3.125e+33, 1.5625e+40], ('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$'], ('timestamp', 'datetime64[ns]'): [Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00')], ('weight(t)', 'float64'): [4.3, 2.0, 4.0, 1.8, 5.7, nan]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_cumprod_numeric(self):
        df = self.df.copy()
        result = df.cols.cumprod(cols='weight(t)')
        expected = self.create_dataframe(data={('NullType', 'object'): [None, None, None, None, None, None], ('attributes', 'object'): [[8.5344, 4300.0], [5.334, 2000.0], [7.9248, 4000.0], [3.9624, 1800.0], [None, 5700.0], [91.44, None]], ('date arrival', 'object'): ['1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10'], ('function(binary)', 'object'): [bytearray(b'Leader'), bytearray(b'Espionage'), bytearray(b'Security'), bytearray(b'First Lieutenant'), bytearray(b'None'), bytearray(b'Battle Station')], ('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 13.0, nan, 300.0], ('japanese name', 'object'): [['Inochi', 'Convoy'], ['Bumble', 'Goldback'], ['Roadbuster'], ['Meister'], ['Megatron'], ['Metroflex']], ('last date seen', 'object'): ['2016/09/10', '2015/08/10', '2014/07/10', '2013/06/10', '2012/05/10', '2011/04/10'], ('last position seen', 'object'): ['19.442735,-99.201111', '10.642707,-71.612534', '37.789563,-122.400356', '33.670666,-117.841553', None, None], ('rank', 'int64'): [10, 7, 7, 8, 10, 8], ('Cybertronian', 'bool'): [True, True, True, True, True, False], ('Date Type', 'datetime64[ns]'): [Timestamp('2016-09-10 00:00:00'), Timestamp('2015-08-10 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2013-06-24 00:00:00'), Timestamp('2012-05-10 00:00:00'), Timestamp('2011-04-10 00:00:00')], ('age', 'int64'): [5000000, 5000000, 5000000, 5000000, 5000000, 5000000], ('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$'], ('timestamp', 'datetime64[ns]'): [Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00')], ('weight(t)', 'float64'): [4.3, 8.6, 34.4, 61.92, 352.944, nan]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_cumsum_all(self):
        df = self.df.copy()
        result = df.cols.cumsum(cols='*')
        expected = self.create_dataframe(data={('NullType', 'float64'): [nan, nan, nan, nan, nan, nan], ('attributes', 'float64'): [nan, nan, nan, nan, nan, nan], ('date arrival', 'float64'): [nan, nan, nan, nan, nan, nan], ('function(binary)', 'float64'): [nan, nan, nan, nan, nan, nan], ('height(ft)', 'float64'): [-28.0, -11.0, 15.0, 28.0, nan, 328.0], ('japanese name', 'float64'): [nan, nan, nan, nan, nan, nan], ('last date seen', 'float64'): [nan, nan, nan, nan, nan, nan], ('last position seen', 'float64'): [nan, nan, nan, nan, nan, nan], ('rank', 'float64'): [10.0, 17.0, 24.0, 32.0, 42.0, 50.0], ('Cybertronian', 'float64'): [1.0, 2.0, 3.0, 4.0, 5.0, 5.0], ('Date Type', 'float64'): [1.4734656e+18, 2.9126304e+18, 4.3161984e+18, 5.6882304e+18, 7.0248384e+18, 8.327232e+18], ('age', 'float64'): [5000000.0, 10000000.0, 15000000.0, 20000000.0, 25000000.0, 30000000.0], ('function', 'float64'): [nan, nan, nan, nan, nan, nan], ('names', 'float64'): [nan, nan, nan, nan, nan, nan], ('timestamp', 'float64'): [1.403568e+18, 2.807136e+18, 4.210704e+18, 5.614272e+18, 7.01784e+18, 8.421408e+18], ('weight(t)', 'float64'): [4.3, 6.3, 10.3, 12.100000000000001, 17.8, nan]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_cumsum_multiple(self):
        df = self.df.copy()
        result = df.cols.cumsum(cols=['height(ft)', 'age', 'rank'])
        expected = self.create_dataframe(data={('NullType', 'object'): [None, None, None, None, None, None], ('attributes', 'object'): [[8.5344, 4300.0], [5.334, 2000.0], [7.9248, 4000.0], [3.9624, 1800.0], [None, 5700.0], [91.44, None]], ('date arrival', 'object'): ['1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10'], ('function(binary)', 'object'): [bytearray(b'Leader'), bytearray(b'Espionage'), bytearray(b'Security'), bytearray(b'First Lieutenant'), bytearray(b'None'), bytearray(b'Battle Station')], ('height(ft)', 'float64'): [-28.0, -11.0, 15.0, 28.0, nan, 328.0], ('japanese name', 'object'): [['Inochi', 'Convoy'], ['Bumble', 'Goldback'], ['Roadbuster'], ['Meister'], ['Megatron'], ['Metroflex']], ('last date seen', 'object'): ['2016/09/10', '2015/08/10', '2014/07/10', '2013/06/10', '2012/05/10', '2011/04/10'], ('last position seen', 'object'): ['19.442735,-99.201111', '10.642707,-71.612534', '37.789563,-122.400356', '33.670666,-117.841553', None, None], ('rank', 'float64'): [10.0, 17.0, 24.0, 32.0, 42.0, 50.0], ('Cybertronian', 'bool'): [True, True, True, True, True, False], ('Date Type', 'datetime64[ns]'): [Timestamp('2016-09-10 00:00:00'), Timestamp('2015-08-10 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2013-06-24 00:00:00'), Timestamp('2012-05-10 00:00:00'), Timestamp('2011-04-10 00:00:00')], ('age', 'float64'): [5000000.0, 10000000.0, 15000000.0, 20000000.0, 25000000.0, 30000000.0], ('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$'], ('timestamp', 'datetime64[ns]'): [Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00')], ('weight(t)', 'float64'): [4.3, 2.0, 4.0, 1.8, 5.7, nan]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_cumsum_numeric(self):
        df = self.df.copy()
        result = df.cols.cumsum(cols='weight(t)')
        expected = self.create_dataframe(data={('NullType', 'object'): [None, None, None, None, None, None], ('attributes', 'object'): [[8.5344, 4300.0], [5.334, 2000.0], [7.9248, 4000.0], [3.9624, 1800.0], [None, 5700.0], [91.44, None]], ('date arrival', 'object'): ['1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10'], ('function(binary)', 'object'): [bytearray(b'Leader'), bytearray(b'Espionage'), bytearray(b'Security'), bytearray(b'First Lieutenant'), bytearray(b'None'), bytearray(b'Battle Station')], ('height(ft)', 'float64'): [-28.0, 17.0, 26.0, 13.0, nan, 300.0], ('japanese name', 'object'): [['Inochi', 'Convoy'], ['Bumble', 'Goldback'], ['Roadbuster'], ['Meister'], ['Megatron'], ['Metroflex']], ('last date seen', 'object'): ['2016/09/10', '2015/08/10', '2014/07/10', '2013/06/10', '2012/05/10', '2011/04/10'], ('last position seen', 'object'): ['19.442735,-99.201111', '10.642707,-71.612534', '37.789563,-122.400356', '33.670666,-117.841553', None, None], ('rank', 'int64'): [10, 7, 7, 8, 10, 8], ('Cybertronian', 'bool'): [True, True, True, True, True, False], ('Date Type', 'datetime64[ns]'): [Timestamp('2016-09-10 00:00:00'), Timestamp('2015-08-10 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2013-06-24 00:00:00'), Timestamp('2012-05-10 00:00:00'), Timestamp('2011-04-10 00:00:00')], ('age', 'int64'): [5000000, 5000000, 5000000, 5000000, 5000000, 5000000], ('function', 'object'): ['Leader', 'Espionage', 'Security', 'First Lieutenant', 'None', 'Battle Station'], ('names', 'object'): ['Optimus', 'bumbl#ebéé  ', 'ironhide&', 'Jazz', 'Megatron', 'Metroplex_)^$'], ('timestamp', 'datetime64[ns]'): [Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00'), Timestamp('2014-06-24 00:00:00')], ('weight(t)', 'float64'): [4.3, 6.3, 10.3, 12.100000000000001, 17.8, nan]}, force_data_types=True)
        self.assertTrue(result.equals(expected, decimal=True, assertion=True))

    def test_cols_kurtosis_all(self):
        df = self.df.copy()
        result = df.cols.kurtosis(cols='*')
        expected = {'NullType': nan, 'attributes': nan, 'date arrival': nan, 'function(binary)': nan, 'height(ft)': 0.1386336481508219, 'japanese name': nan, 'last date seen': nan, 'last position seen': nan, 'rank': -1.5000000000000002, 'Cybertronian': 1.200000000000001, 'Date Type': -1.2534316806077568, 'age': -3.0, 'function': nan, 'names': nan, 'timestamp': -3.0, 'weight(t)': -1.4364057698737658}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_kurtosis_multiple(self):
        df = self.df.copy()
        result = df.cols.kurtosis(cols=['height(ft)', 'age', 'rank'])
        expected = {'height(ft)': 0.1386336481508219, 'age': -3.0, 'rank': -1.5000000000000002}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_kurtosis_numeric(self):
        df = self.df.copy()
        result = df.cols.kurtosis(cols='weight(t)')
        expected = -1.4364057698737658
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_mad_all(self):
        df = self.df.copy()
        result = df.cols.mad(cols=['NullType', 'attributes', 'date arrival', 'function(binary)', 'height(ft)', 'japanese name', 'last date seen', 'last position seen', 'rank', 'Cybertronian', 'age', 'function', 'names', 'timestamp', 'weight(t)'], estimate=False)
        expected = {'NullType': nan, 'attributes': nan, 'date arrival': nan, 'function(binary)': nan, 'height(ft)': 9.0, 'japanese name': nan, 'last date seen': nan, 'last position seen': nan, 'rank': 1.0, 'Cybertronian': 0.0, 'age': 0.0, 'function': nan, 'names': nan, 'timestamp': 0.0, 'weight(t)': 1.7000000000000002}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_mad_multiple(self):
        df = self.df.copy()
        result = df.cols.mad(cols=['height(ft)', 'age', 'rank'], more=True, estimate=False)
        expected = {'height(ft)': {'mad': 9.0, 'median': 17.0}, 'age': {'mad': 0.0, 'median': 5000000.0}, 'rank': {'mad': 1.0, 'median': 8.0}}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_mad_numeric(self):
        df = self.df.copy()
        result = df.cols.mad(cols='weight(t)', relative_error=0.45, estimate=False)
        expected = 1.7000000000000002
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_max_all(self):
        df = self.df.copy()
        result = df.cols.max(cols='*')
        expected = {'NullType': nan, 'attributes': '[None, 5700.0]', 'date arrival': '1980/04/10', 'function(binary)': "bytearray(b'Security')", 'height(ft)': 300.0, 'japanese name': "['Roadbuster']", 'last date seen': '2016/09/10', 'last position seen': '37.789563,-122.400356', 'rank': 10, 'Cybertronian': True, 'Date Type': datetime.datetime(2016, 9, 10, 0, 0), 'age': 5000000, 'function': 'Security', 'names': 'ironhide&', 'timestamp': datetime.datetime(2014, 6, 24, 0, 0), 'weight(t)': 5.7}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_max_multiple(self):
        df = self.df.copy()
        result = df.cols.max(cols=['height(ft)', 'age', 'rank'])
        expected = {'height(ft)': 300.0, 'age': 5000000.0, 'rank': 10.0}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_max_numeric(self):
        df = self.df.copy()
        result = df.cols.max(cols='weight(t)')
        expected = 5.7
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_mean_all(self):
        df = self.df.copy()
        result = df.cols.mean(cols='*')
        expected = {'NullType': nan, 'attributes': nan, 'date arrival': nan, 'function(binary)': nan, 'height(ft)': 65.6, 'japanese name': nan, 'last date seen': nan, 'last position seen': nan, 'rank': 8.333333333333334, 'Cybertronian': 0.8333333333333334, 'Date Type': 1.387872e+18, 'age': 5000000.0, 'function': nan, 'names': nan, 'timestamp': 1.403568e+18, 'weight(t)': 3.56}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_mean_multiple(self):
        df = self.df.copy()
        result = df.cols.mean(cols=['height(ft)', 'age', 'rank'])
        expected = {'height(ft)': 65.6, 'age': 5000000.0, 'rank': 8.333333333333334}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_mean_numeric(self):
        df = self.df.copy()
        result = df.cols.mean(cols='weight(t)')
        expected = 3.56
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_median_all(self):
        df = self.df.copy()
        result = df.cols.median(cols='*')
        expected = {'NullType': nan, 'attributes': nan, 'date arrival': nan, 'function(binary)': nan, 'height(ft)': {0.5: 17.0}, 'japanese name': nan, 'last date seen': nan, 'last position seen': nan, 'rank': {0.5: 8.0}, 'Cybertronian': {0.5: 1.0}, 'Date Type': {0.5: 1.3878e+18}, 'age': {0.5: 5000000.0}, 'function': nan, 'names': nan, 'timestamp': {0.5: 1.403568e+18}, 'weight(t)': {0.5: 4.0}}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_median_multiple(self):
        df = self.df.copy()
        result = df.cols.median(cols=['height(ft)', 'age', 'rank'], relative_error=0.012)
        expected = {'height(ft)': {0.5: 17.0}, 'age': {0.5: 5000000.0}, 'rank': {0.5: 8.0}}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_median_numeric(self):
        df = self.df.copy()
        result = df.cols.median(cols='weight(t)', relative_error=0.33)
        expected = None
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_min_all(self):
        df = self.df.copy()
        result = df.cols.min(cols='*')
        expected = {'NullType': nan, 'attributes': '[3.9624, 1800.0]', 'date arrival': '1980/04/10', 'function(binary)': "bytearray(b'Battle Station')", 'height(ft)': -28.0, 'japanese name': "['Bumble', 'Goldback']", 'last date seen': '2011/04/10', 'last position seen': '10.642707,-71.612534', 'rank': 7, 'Cybertronian': False, 'Date Type': datetime.datetime(2011, 4, 10, 0, 0), 'age': 5000000, 'function': 'Battle Station', 'names': 'Jazz', 'timestamp': datetime.datetime(2014, 6, 24, 0, 0), 'weight(t)': 1.8}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_min_multiple(self):
        df = self.df.copy()
        result = df.cols.min(cols=['height(ft)', 'age', 'rank'])
        expected = {'height(ft)': -28.0, 'age': 5000000.0, 'rank': 7.0}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_min_numeric(self):
        df = self.df.copy()
        result = df.cols.min(cols='weight(t)')
        expected = 1.8
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_mode_all(self):
        df = self.create_dataframe(data={('weight', 'float64'): [10.2, 10.2, 20.4, 20.5, 33.2, nan], ('height', 'float64'): [26.0, 17.0, 26.0, 17.0, nan, 300.0], ('date arrival', 'object'): ['1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', nan], ('age', 'float64'): [5000000.0, 5000000.0, 5000000.0, 5000000.0, 5000000.0, nan], ('rank', 'int64'): [10, 7, 7, 8, 10, 8]}, force_data_types=True)
        result = df.cols.mode(cols='*')
        expected = {'weight': 10.2, 'height': [17.0, 26.0], 'date arrival': '1980/04/10', 'age': 5000000.0, 'rank': [7, 8, 10]}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_mode_multiple(self):
        df = self.create_dataframe(data={('weight', 'float64'): [10.2, 10.2, 20.4, 20.5, 33.2, nan], ('height', 'float64'): [26.0, 17.0, 26.0, 17.0, nan, 300.0], ('date arrival', 'object'): ['1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', nan], ('age', 'float64'): [5000000.0, 5000000.0, 5000000.0, 5000000.0, 5000000.0, nan], ('rank', 'int64'): [10, 7, 7, 8, 10, 8]}, force_data_types=True)
        result = df.cols.mode(cols=['height', 'age', 'rank'])
        expected = {'height': [17.0, 26.0], 'age': 5000000.0, 'rank': [7, 8, 10]}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_mode_numeric(self):
        df = self.create_dataframe(data={('weight', 'float64'): [10.2, 10.2, 20.4, 20.5, 33.2, nan], ('height', 'float64'): [26.0, 17.0, 26.0, 17.0, nan, 300.0], ('date arrival', 'object'): ['1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', '1980/04/10', nan], ('age', 'float64'): [5000000.0, 5000000.0, 5000000.0, 5000000.0, 5000000.0, nan], ('rank', 'int64'): [10, 7, 7, 8, 10, 8]}, force_data_types=True)
        result = df.cols.mode(cols='weight')
        expected = 10.2
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_percentile_all(self):
        df = self.df.copy()
        result = df.cols.percentile(cols='*', estimate=False)
        expected = {'NullType': nan, 'attributes': nan, 'date arrival': nan, 'function(binary)': nan, 'height(ft)': {0.25: 13.0, 0.5: 17.0, 0.75: 26.0}, 'japanese name': nan, 'last date seen': nan, 'last position seen': nan, 'rank': {0.25: 7.25, 0.5: 8.0, 0.75: 9.5}, 'Cybertronian': {0.25: 1.0, 0.5: 1.0, 0.75: 1.0}, 'Date Type': {0.25: 1.345464e+18, 0.5: 1.3878e+18, 0.75: 1.4302656e+18}, 'age': {0.25: 5000000.0, 0.5: 5000000.0, 0.75: 5000000.0}, 'function': nan, 'names': nan, 'timestamp': {0.25: 1.403568e+18, 0.5: 1.403568e+18, 0.75: 1.403568e+18}, 'weight(t)': {0.25: 2.0, 0.5: 4.0, 0.75: 4.3}}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_percentile_multiple(self):
        df = self.df.copy()
        result = df.cols.percentile(cols=['height(ft)', 'age', 'rank'], values=0.95, relative_error=0.012, estimate=False)
        expected = {'height(ft)': 245.19999999999996, 'age': 5000000.0, 'rank': 10.0}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_percentile_numeric_multiple(self):
        df = self.df.copy()
        result = df.cols.percentile(cols='weight(t)', values=[0.25, 0.5, 0.75], estimate=False)
        expected = {0.25: 2.0, 0.5: 4.0, 0.75: 4.3}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_percentile_numeric_single(self):
        df = self.df.copy()
        result = df.cols.percentile(cols='weight(t)', values=0.3, estimate=False)
        expected = 2.4
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_range_all(self):
        df = self.df.copy()
        result = df.cols.range(cols='*')
        expected = {'NullType': {'min': nan, 'max': nan}, 'attributes': {'min': nan, 'max': nan}, 'date arrival': {'min': nan, 'max': nan}, 'function(binary)': {'min': nan, 'max': nan}, 'height(ft)': {'min': -28.0, 'max': 300.0}, 'japanese name': {'min': nan, 'max': nan}, 'last date seen': {'min': nan, 'max': nan}, 'last position seen': {'min': nan, 'max': nan}, 'rank': {'min': 7.0, 'max': 10.0}, 'Cybertronian': {'min': 0.0, 'max': 1.0}, 'Date Type': {'min': 1.3023936e+18, 'max': 1.4734656e+18}, 'age': {'min': 5000000.0, 'max': 5000000.0}, 'function': {'min': nan, 'max': nan}, 'names': {'min': nan, 'max': nan}, 'timestamp': {'min': 1.403568e+18, 'max': 1.403568e+18}, 'weight(t)': {'min': 1.8, 'max': 5.7}}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_range_multiple(self):
        df = self.df.copy()
        result = df.cols.range(cols=['height(ft)', 'age', 'rank'])
        expected = {'height(ft)': {'min': -28.0, 'max': 300.0}, 'age': {'min': 5000000.0, 'max': 5000000.0}, 'rank': {'min': 7.0, 'max': 10.0}}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_range_numeric(self):
        df = self.df.copy()
        result = df.cols.range(cols='weight(t)')
        expected = {'min': 1.8, 'max': 5.7}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_skew_all(self):
        df = self.df.copy()
        result = df.cols.skew(cols='*')
        expected = {'NullType': nan, 'attributes': nan, 'date arrival': nan, 'function(binary)': nan, 'height(ft)': 1.4048993602366326, 'japanese name': nan, 'last date seen': nan, 'last position seen': nan, 'rank': 0.38180177416060485, 'Cybertronian': -1.7888543819998322, 'Date Type': 0.002228190425762595, 'age': 0.0, 'function': nan, 'names': nan, 'timestamp': 0.0, 'weight(t)': 0.06521107849040725}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_skew_multiple(self):
        df = self.df.copy()
        result = df.cols.skew(cols=['height(ft)', 'age', 'rank'])
        expected = {'height(ft)': 1.4048993602366326, 'age': 0.0, 'rank': 0.38180177416060485}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_skew_numeric(self):
        df = self.df.copy()
        result = df.cols.skew(cols='weight(t)')
        expected = 0.06521107849040725
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_std_all(self):
        df = self.df.copy()
        result = df.cols.std(cols=['NullType', 'attributes', 'date arrival', 'function(binary)', 'height(ft)', 'japanese name', 'last date seen', 'last position seen', 'rank', 'Cybertronian', 'age', 'function', 'names', 'timestamp', 'weight(t)'])
        expected = {'NullType': nan, 'attributes': nan, 'date arrival': nan, 'function(binary)': nan, 'height(ft)': 132.66612227694, 'japanese name': nan, 'last date seen': nan, 'last position seen': nan, 'rank': 1.3662601021279464, 'Cybertronian': 0.408248290463863, 'age': 0.0, 'function': nan, 'names': nan, 'timestamp': 0.0, 'weight(t)': 1.6471186963907611}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_std_multiple(self):
        df = self.df.copy()
        result = df.cols.std(cols=['height(ft)', 'age', 'rank'])
        expected = {'height(ft)': 132.66612227694, 'age': 0.0, 'rank': 1.3662601021279464}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_std_numeric(self):
        df = self.df.copy()
        result = df.cols.std(cols='weight(t)')
        expected = 1.6471186963907611
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_sum_all(self):
        df = self.df.copy()
        result = df.cols.sum(cols='*')
        expected = {'NullType': 0.0, 'attributes': 0.0, 'date arrival': 0.0, 'function(binary)': 0.0, 'height(ft)': 328.0, 'japanese name': 0.0, 'last date seen': 0.0, 'last position seen': 0.0, 'rank': 50.0, 'Cybertronian': 5.0, 'Date Type': 8.327232e+18, 'age': 30000000.0, 'function': 0.0, 'names': 0.0, 'timestamp': 8.421408e+18, 'weight(t)': 17.8}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_sum_multiple(self):
        df = self.df.copy()
        result = df.cols.sum(cols=['height(ft)', 'age', 'rank'])
        expected = {'height(ft)': 328.0, 'age': 30000000.0, 'rank': 50.0}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_sum_numeric(self):
        df = self.df.copy()
        result = df.cols.sum(cols='weight(t)')
        expected = 17.8
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_var_all(self):
        df = self.df.copy()
        result = df.cols.var(cols=['NullType', 'attributes', 'date arrival', 'function(binary)', 'height(ft)', 'japanese name', 'last date seen', 'last position seen', 'rank', 'Cybertronian', 'age', 'function', 'names', 'timestamp', 'weight(t)'])
        expected = {'NullType': nan, 'attributes': nan, 'date arrival': nan, 'function(binary)': nan, 'height(ft)': 17600.3, 'japanese name': nan, 'last date seen': nan, 'last position seen': nan, 'rank': 1.8666666666666665, 'Cybertronian': 0.16666666666666669, 'age': 0.0, 'function': nan, 'names': nan, 'timestamp': 0.0, 'weight(t)': 2.713}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_var_multiple(self):
        df = self.df.copy()
        result = df.cols.var(cols=['height(ft)', 'age', 'rank'])
        expected = {'height(ft)': 17600.3, 'age': 0.0, 'rank': 1.8666666666666665}
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))

    def test_cols_var_numeric(self):
        df = self.df.copy()
        result = df.cols.var(cols='weight(t)')
        expected = 2.713
        self.assertTrue(results_equal(result, expected, decimal=5, assertion=True))


class TestStatisticsDask(TestStatisticsPandas):
    config = {'engine': 'dask', 'n_partitions': 1}


class TestStatisticsPartitionDask(TestStatisticsPandas):
    config = {'engine': 'dask', 'n_partitions': 2}


try:
    import cudf # pyright: reportMissingImports=false
except:
    pass
else:
    class TestStatisticsCUDF(TestStatisticsPandas):
        config = {'engine': 'cudf'}


try:
    import dask_cudf # pyright: reportMissingImports=false
except:
    pass
else:
    class TestStatisticsDC(TestStatisticsPandas):
        config = {'engine': 'dask_cudf', 'n_partitions': 1}


try:
    import dask_cudf # pyright: reportMissingImports=false
except:
    pass
else:
    class TestStatisticsPartitionDC(TestStatisticsPandas):
        config = {'engine': 'dask_cudf', 'n_partitions': 2}


try:
    import pyspark # pyright: reportMissingImports=false
except:
    pass
else:
    class TestStatisticsSpark(TestStatisticsPandas):
        config = {'engine': 'spark'}


try:
    import vaex # pyright: reportMissingImports=false
except:
    pass
else:
    class TestStatisticsVaex(TestStatisticsPandas):
        config = {'engine': 'vaex'}



from abc import ABC
from functools import reduce
from typing import Callable

from optimus.engines.base.meta import Meta
from optimus.helpers.columns import parse_columns, prepare_columns_arguments
from optimus.helpers.constants import Actions
from optimus.helpers.core import val_to_list
from optimus.helpers.raiseit import RaiseIt
from optimus.helpers.types import *
from optimus.infer import is_dict, is_list_of_str, is_list_of_tuples, is_list_value, is_str


class BaseRows(ABC):
    """Base class for all Rows implementations"""

    def __init__(self, root: 'DataFrameType'):
        self.root = root

    def append(self, dfs: 'DataFrameTypeList', names_map=None) -> 'DataFrameType':
        """
        Appends 2 or more dataframes
        :param dfs:
        :param names_map:
        """
        if not is_list_value(dfs):
            dfs = [dfs]

        every_df = [self.root, *dfs]

        if names_map is not None:
            rename: list[list] = [[] for _ in every_df]

            for key in names_map:
                assert len(names_map[key]) == len(every_df)
                for i, col_name in enumerate(names_map[key]):
                    if col_name:
                        rename[i] = [*rename[i], (col_name, "__output_column__" + key)]

            for i, _rename in enumerate(rename):
                every_df[i] = every_df[i].cols.rename(_rename)
                select = [output_name for name, output_name in _rename]
                every_df[i] = every_df[i].cols.select(select)

        dfd = every_df[0].data

        for i, _df in enumerate(every_df):
            if i != 0:
                dfd = self.root.functions.append(dfd, _df.data)

        df = self.root.new(dfd)

        if names_map is not None:
            df = df.cols.rename([("__output_column__" + key, key) for key in names_map])
            df = df.cols.select([*names_map.keys()])

        return df.new(df.data.reset_index(drop=True))

    def apply(self, func, args=None, output_cols=None, mode="vectorized", **kwargs) -> 'DataFrameType':
        """
        Apply a function to each row of the dataframe
        :param func: function to be applied
        :param args: list of arguments to pass to 'func'
        :param output_cols: pass the result to one or more columns, if None, the
        result will replace the original dataframe
        :param mode: 
        :return:
        """
        df = self.root
        dfd = self.root.data
        kw_columns = {}
        args = args or ()

        if output_cols:
            output_cols = val_to_list(output_cols)
            func = prepare_columns_arguments(output_cols, func)

        if mode == "vectorized":
            if output_cols is None:
                data = func(dfd, *args, **kwargs)
            for output_col, _func in zip(output_cols, func):
                result = _func(dfd, *args, **kwargs)
                kw_columns = {output_col: result}

        elif mode == "map":
            if output_cols is None:
                data = dfd.apply(func, axis=1, args=args, **kwargs)
            else:
                for output_col, _func in zip(output_cols, func):
                    result = _func(dfd, *args, **kwargs)
                    kw_columns = {output_col: result}
        else:
            RaiseIt.value_error(mode, ["map", "vectorized"])

        if output_cols is None:
            df = self.root.op.create.dataframe(data)
        else:
            df = df.cols.assign(kw_columns)

        return df

    def select(self, expr=None, contains=None, case=None, flags=0, na=False, regex=False) -> 'DataFrameType':
        """
        Return selected rows using an expression
        :param expr: Expression used, For Ex: (df["A"] > 3) & (df["A"] <= 1000) or Column name "A"
        :param contains: List of string
        :param case:
        :param flags:
        :param na:
        :param regex:
        :return:
        """

        df = self.root
        dfd = df.data

        all_cols = df.cols.names()

        if is_str(expr) and expr == "*":
            expr = all_cols

        expr = val_to_list(expr)

        if is_list_of_str(expr) and all([col in all_cols for col in expr]):
            if contains is not None:
                expr = [df.mask.contains(expr, value=contains, case=case, flags=flags, na=na, regex=regex).mask.any()]
            else:
                expr = [df[expr].mask.any()]
        else:
            for i, expr_ in enumerate(expr):
                if not is_str(expr_):
                    continue
                elif expr_ not in all_cols:
                    expr[i] = eval(expr_)
                elif contains is not None:
                    expr[i] = df.mask.contains(expr_, value=contains, case=case, flags=flags, na=na, regex=regex)
                else:
                    expr[i] = df[expr_]

        if len(expr) > 1:
            def _reduce_exprs(a, b):
                if hasattr(a, "get_series"):
                    a = a.get_series()
                if hasattr(b, "get_series"):
                    b = b.get_series()
                return df.functions.to_boolean(a) & df.functions.to_boolean(b)

            expr = reduce(_reduce_exprs, expr)
        else:
            expr = expr[0]

            if hasattr(expr, "get_series"):
                expr = expr.get_series()

            expr = df.functions.to_boolean(expr)

        dfd = dfd.reset_index(drop=True)[expr.reset_index(drop=True)].reset_index(drop=True)

        meta = Meta.action(df.meta, Actions.SELECT_ROW.value, df.cols.names())

        df = self.root.new(dfd, meta=meta)
        return df

    def _count(self, compute=True) -> int:
        """
        Specific implementation to count the number of rows in the dataframe
        :param compute:
        :return:
        """
        return len(self.root.data.index)

    def count(self, compute=True) -> int:
        """
        Count dataframe rows
        """
        df = self.root
        dfd = df.data
        # TODO: Be sure that we need the compute param
        if compute is True:
            result = self._count(compute)
        else:
            result = df.functions.delayed(len)(dfd)
        return result

    def to_list(self, input_cols) -> list:
        """

        :param input_cols:
        :return:
        """
        df = self.root
        input_cols = parse_columns(df, input_cols)
        value = df.cols.select(input_cols).to_pandas().values.tolist()

        return value

    def sort(self, cols="*", order="desc", cast=True) -> 'DataFrameType':
        """
        Sort rows taking into account multiple columns
        :param cols:
        :param order:
        :param cast: cast rows before sorting them.
        """
        df = self.root

        if is_dict(cols):
            cols = list(cols.items())

        if is_list_of_tuples(cols):
            cols, order = zip(*cols)

        cols = parse_columns(df, cols)
        order = prepare_columns_arguments(cols, order)

        def _set_order(o):
            if is_str(o):
                if o not in ["asc", "desc"]:
                    RaiseIt.value_error(o, ["asc", "desc"])
                o = True if o == "asc" else False
            return o

        order = [_set_order(o) for o in order]

        if cast:
            sort_cols = [f"__{col}_sort__" for col in cols]
            types = df.cols.inferred_data_type(cols, use_internal=True, tidy=False)["inferred_data_type"]
            casts = {}

            for col_name, data_type in types.items():
                _cast = "float" if data_type in df.constants.NUMERIC_INTERNAL_TYPES else "str"
                casts.update({col_name: _cast})

            df = df.cols.cast(casts, output_cols=sort_cols)
        else:
            sort_cols = cols

        dfd = df.functions.sort_df(df.data, sort_cols, order)
        meta = Meta.action(self.root.meta, Actions.SORT_ROW.value, cols)

        if cast:
            dfd = dfd.drop(sort_cols, axis=1)

        df = self.root.new(dfd, meta=meta)

        return df

    def reverse(self) -> 'DataFrameType':
        """

        :return:
        """
        dfd = self.root.functions.reverse_df(self.root.data)
        return self.root.new(dfd)

    def drop(self, where) -> 'DataFrameType':
        """
        Drop rows depending on a mask or an expression
        :param where: Mask, expression or name of the column to be taken as mask
        :return: Optimus Dataframe
        """
        df = self.root
        dfd = df.data

        if is_str(where):
            if where in df.cols.names():
                where = df[where]
            else:
                where = eval(where)
        dfd = dfd.reset_index(drop=True)[where.get_series().reset_index(drop=True) == 0]
        meta = Meta.action(df.meta, Actions.DROP_ROW.value, df.cols.names())
        return self.root.new(dfd, meta=meta)

    def between_index(self, lower_bound=None, upper_bound=None, cols="*"):
        """

        :param columns:
        :param lower_bound:
        :param upper_bound:
        :return:
        """
        df = self.root
        dfd = df.data
        cols = parse_columns(df, cols)
        return self.root.new(dfd[lower_bound: upper_bound][cols])

    def limit(self, count=10) -> 'DataFrameType':
        """
        Limit the number of rows
        :param count:
        :return:
        """
        # Ensure that count is an integer. Ibis complains otherwise
        count = int(count)
        return self.root.new(self.root.data[:count])

    @staticmethod
    def unnest(cols) -> 'DataFrameType':
        """

        :param cols:
        :return:
        """
        raise NotImplementedError('Not implemented yet')

    def approx_count(self) -> 'DataFrameType':
        """
        Aprox count
        :return:
        """
        return self.root.rows.count()

    def _mask(self, cols, func: Callable, drop=False, how="any", *args, **kwargs) -> 'DataFrameType':
        """

        :param cols:
        :param method:
        :param drop:
        :param how:
        :param args:
        :param kwargs:
        :return:
        """
        df = self.root
        mask = func(cols=cols, *args, **kwargs)

        if how == "any":
            mask = mask.mask.any()
        elif how == "all":
            mask = mask.mask.all()
        else:
            RaiseIt.value_error(how, ["any", "all"])

        if drop:
            mask = ~mask

        df = df.rows.select(mask)

        return df

    def str(self, cols="*", drop=False, how="any") -> 'DataFrameType':
        """
        #TODO:?
        :param cols:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.str, drop=drop, how=how)

    def int(self, cols="*", drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.int, drop=drop, how=how)

    def float(self, cols="*", drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.float, drop=drop, how=how)

    def numeric(self, cols="*", drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.numeric, drop=drop, how=how)

    def between(self, cols="*", lower_bound=None, upper_bound=None, equal=True, bounds=None, drop=False,
                how="any") -> 'DataFrameType':
        """

        :param cols:
        :param lower_bound:
        :param upper_bound:
        :param equal:
        :param bounds:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.between, drop=drop, how=how, lower_bound=lower_bound,
                          upper_bound=upper_bound, equal=equal, bounds=bounds)

    def greater_than_equal(self, cols="*", value=None, drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param value:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.greater_than_equal, drop=drop, value=value, how=how)

    def greater_than(self, cols="*", value=None, drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param value:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.greater_than, drop=drop, value=value, how=how)

    def less_than(self, cols="*", value=None, drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param value:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.less_than, drop=drop, value=value, how=how)

    def less_than_equal(self, cols="*", value=None, drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param value:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.less_than_equal, drop=drop, value=value, how=how)

    def equal(self, cols="*", value=None, drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param value:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.equal, drop=drop, value=value, how=how)

    def not_equal(self, cols="*", value=None, drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param value:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.not_equal, drop=drop, value=value, how=how)

    def missing(self, cols="*", drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.missing, drop=drop, how=how)

    def null(self, cols="*", drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.null, drop=drop, how=how)

    def none(self, cols="*", drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.none, drop=drop, how=how)

    def nan(self, cols="*", drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.nan, drop=drop, how=how)

    def empty(self, cols="*", drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.empty, drop=drop, how=how)

    def duplicated(self, cols="*", keep="first", drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param keep:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.duplicated, drop=drop, keep=keep, how=how)

    def unique(self, cols="*", keep="first", drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param keep:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.unique, drop=drop, keep=keep, how=how)

    def mismatch(self, cols="*", data_type=None, drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param data_type:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.mismatch, drop=drop, data_type=data_type, how=how)

    def match(self, cols="*", regex=None, data_type=None, drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param regex:
        :param data_type:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.match, drop=drop, regex=regex, data_type=data_type, how=how)

    def match_regex(self, cols="*", regex=None, drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param regex:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.match_regex, drop=drop, regex=regex, how=how)

    def match_data_type(self, cols="*", data_type=None, drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param data_type:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.match_data_type, drop=drop, data_type=data_type, how=how)

    def value_in(self, cols="*", values=None, drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param values:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.value_in, drop=drop, values=values, how=how)

    def pattern(self, cols="*", pattern=None, drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param pattern:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.pattern, drop=drop, pattern=pattern, how=how)

    def starts_with(self, cols="*", value=None, drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param value:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.starts_with, drop=drop, value=value, how=how)

    def ends_with(self, cols="*", value=None, drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param value:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.ends_with, drop=drop, value=value, how=how)

    def contains(self, cols="*", value=None, drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param value:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.contains, drop=drop, value=value, how=how)

    def find(self, cols="*", value=None, drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param value:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.find, drop=drop, value=value, how=how)

    def email(self, cols="*", drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.email, drop=drop, how=how)

    def ip(self, cols="*", drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.ipv4, drop=drop, how=how)

    def url(self, cols="*", drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.url, drop=drop, how=how)

    def gender(self, cols="*", drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.gender, drop=drop, how=how)

    def boolean(self, cols="*", drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.boolean, drop=drop, how=how)

    def zip_code(self, cols="*", drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.zip_code, drop=drop, how=how)

    def credit_card_number(self, cols="*", drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.credit_card_number, drop=drop, how=how)

    def datetime(self, cols="*", drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.datetime, drop=drop, how=how)

    def object(self, cols="*", drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.object, drop=drop, how=how)

    def array(self, cols="*", drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.array, drop=drop, how=how)

    def phone_number(self, cols="*", drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.phone_number, drop=drop, how=how)

    def social_security_number(self, cols="*", drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.social_security_number, drop=drop, how=how)

    def http_code(self, cols="*", drop=False, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.http_code, drop=drop, how=how)

    def expression(self, where=None, cols="*", drop=False, how="any") -> 'DataFrameType':
        """

        :param where:
        :param cols:
        :param drop:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.expression, drop=drop, how=how, where=where)

    # drop functions
    def drop_str(self, cols="*", how="any") -> 'DataFrameType':
        """

        :param cols:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.str, drop=True, how=how)

    def drop_int(self, cols="*", how="any") -> 'DataFrameType':
        """

        :param cols:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.int, drop=True, how=how)

    def drop_float(self, cols="*", how="any") -> 'DataFrameType':
        """

        :param cols:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.float, drop=True, how=how)

    def drop_numeric(self, cols="*", how="any") -> 'DataFrameType':
        """

        :param cols:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.numeric, drop=True, how=how)

    def drop_greater_than_equal(self, cols="*", value=None, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param value:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.greater_than_equal, drop=True, value=value, how=how)

    def drop_greater_than(self, cols="*", value=None, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param value:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.greater_than, drop=True, value=value, how=how)

    def drop_less_than_equal(self, cols="*", value=None, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param value:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.less_than_equal, drop=True, value=value, how=how)

    def drop_less_than(self, cols="*", value=None, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param value:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.less_than, drop=True, value=value, how=how)

    def drop_between(self, cols="*", lower_bound=None, upper_bound=None, equal=True, bounds=None,
                     how="any") -> 'DataFrameType':
        """

        :param cols:
        :param lower_bound:
        :param upper_bound:
        :param equal:
        :param bounds:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.between, drop=True, how=how, lower_bound=lower_bound,
                          upper_bound=upper_bound, equal=equal, bounds=bounds)

    def drop_equal(self, cols="*", value=None, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param value:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.equal, drop=True, value=value, how=how)

    def drop_not_equal(self, cols="*", value=None, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param value:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.not_equal, drop=True, value=value, how=how)

    def drop_missings(self, cols="*", how="any") -> 'DataFrameType':
        """

        :param cols:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.missing, drop=True, how=how)

    def drop_nulls(self, cols="*", how="any") -> 'DataFrameType':
        """

        :param cols:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.null, drop=True, how=how)

    def drop_none(self, cols="*", how="any") -> 'DataFrameType':
        """

        :param cols:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.none, drop=True, how=how)

    def drop_nan(self, cols="*", how="any") -> 'DataFrameType':
        """

        :param cols:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.nan, drop=True, how=how)

    def drop_empty(self, cols="*", how="any") -> 'DataFrameType':
        """

        :param cols:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.empty, drop=True, how=how)

    def drop_duplicated(self, cols="*", keep="first", how="any") -> 'DataFrameType':
        """

        :param cols:
        :param keep:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.duplicated, drop=True, keep=keep, how=how)

    def drop_uniques(self, cols="*", keep="first", how="any") -> 'DataFrameType':
        """
        Drops first (passed to keep) matches of duplicates and unique values.
        :param cols: 
        :param keep: 
        :param how: 
        :return: Dataframe
        """
        return self._mask(cols, func=self.root.mask.unique, drop=True, keep=keep, how=how)

    def drop_mismatch(self, cols="*", data_type=None, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param data_type:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.mismatch, drop=True, data_type=data_type, how=how)

    def drop_match(self, cols="*", regex=None, data_type=None, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param regex:
        :param data_type:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.match, drop=True, regex=regex, data_type=data_type, how=how)

    def drop_by_regex(self, cols="*", regex=None, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param regex:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.match_regex, drop=True, regex=regex, how=how)

    def drop_by_data_type(self, cols="*", data_type=None, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param data_type:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.match_data_type, drop=True, data_type=data_type, how=how)

    def drop_value_in(self, cols="*", values=None, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param values:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.value_in, drop=True, values=values, how=how)

    def drop_pattern(self, cols="*", pattern=None, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param pattern:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.pattern, drop=True, pattern=pattern, how=how)

    def drop_starts_with(self, cols="*", value=None, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param value:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.starts_with, drop=True, value=value, how=how)

    def drop_ends_with(self, cols="*", value=None, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param value:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.ends_with, drop=True, value=value, how=how)

    def drop_contains(self, cols="*", value=None, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param value:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.contains, drop=True, value=value, how=how)

    def drop_find(self, cols="*", value=None, how="any") -> 'DataFrameType':
        """

        :param cols:
        :param value:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.find, drop=True, value=value, how=how)

    def drop_emails(self, cols="*", how="any") -> 'DataFrameType':
        """

        :param cols:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.email, drop=True, how=how)

    def drop_ips(self, cols="*", how="any") -> 'DataFrameType':
        """

        :param cols:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.ipv4, drop=True, how=how)

    def drop_urls(self, cols="*", how="any") -> 'DataFrameType':
        """

        :param cols:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.url, drop=True, how=how)

    def drop_genders(self, cols="*", how="any") -> 'DataFrameType':
        """

        :param cols:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.gender, drop=True, how=how)

    def drop_booleans(self, cols="*", how="any") -> 'DataFrameType':
        """

        :param cols:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.boolean, drop=True, how=how)

    def drop_zip_codes(self, cols="*", how="any") -> 'DataFrameType':
        """

        :param cols:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.zip_code, drop=True, how=how)

    def drop_credit_card_numbers(self, cols="*", how="any") -> 'DataFrameType':
        """

        :param cols:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.credit_card_number, drop=True, how=how)

    def drop_datetimes(self, cols="*", how="any") -> 'DataFrameType':
        """

        :param cols:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.datetime, drop=True, how=how)

    def drop_objects(self, cols="*", how="any") -> 'DataFrameType':
        """

        :param cols:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.object, drop=True, how=how)

    def drop_arrays(self, cols="*", how="any") -> 'DataFrameType':
        """

        :param cols:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.array, drop=True, how=how)

    def drop_phone_numbers(self, cols="*", how="any") -> 'DataFrameType':
        """

        :param cols:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.phone_number, drop=True, how=how)

    def drop_social_security_numbers(self, cols="*", how="any") -> 'DataFrameType':
        """

        :param cols:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.social_security_number, drop=True, how=how)

    def drop_http_codes(self, cols="*", how="any") -> 'DataFrameType':
        """

        :param cols:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.http_code, drop=True, how=how)

    def drop_by_expression(self, where=None, cols="*", how="any") -> 'DataFrameType':
        """

        :param where:
        :param cols:
        :param how:
        :return:
        """
        return self._mask(cols, func=self.root.mask.expression, drop=True, how=how, where=where)



import glob
from optimus.infer import is_url
import uuid
import zipfile
from pathlib import Path
from typing import Tuple

from io import StringIO
import requests

import pandas as pd
import pandavro as pdx


from optimus.optimus import EnginePretty
from optimus.engines.base.io.load import BaseLoad
from optimus.engines.base.meta import Meta
from optimus.engines.pandas.dataframe import PandasDataFrame
from optimus.helpers.functions import prepare_path, unquote_path
from optimus.helpers.logger import logger
from optimus.helpers.core import val_to_list


class Load(BaseLoad):

    @staticmethod
    def df(*args, **kwargs):
        return PandasDataFrame(*args, **kwargs)

    @staticmethod
    def _csv(filepath_or_buffer, *args, **kwargs):
        kwargs.pop("n_partitions", None)

        try:
            # resp = requests.get(filepath_or_buffer)
            df = pd.read_csv(filepath_or_buffer, *args, **kwargs)

        except requests.exceptions.HTTPError as err:
            print(err)

        return df

    @staticmethod
    def _json(filepath_or_buffer, *args, **kwargs):

        def _safe_json(filepath_or_buffer, *args, **kwargs) -> Tuple[pd.DataFrame, bool]:

            try:
                df = pd.read_json(filepath_or_buffer, *args, **kwargs)
                return df, True
            except ValueError:
                kwargs.pop("nrows", None)
                kwargs.pop("lines", None)
                df = pd.read_json(filepath_or_buffer, *args, **kwargs)
                return df, False

        kwargs.pop("n_partitions", None)

        if is_url(filepath_or_buffer):
            s = requests.get(filepath_or_buffer).text
            df, truncated = _safe_json(StringIO(s), *args, **kwargs)
        else:
            df, truncated = _safe_json(filepath_or_buffer, *args, **kwargs)

        if not truncated:
            nrows = kwargs.get("nrows", None)
            if nrows:
                logger.warn(f"'load.json' on {EnginePretty.PANDAS.value} loads the whole dataset and then truncates it if file is not multiline.")
                df = df[:nrows]

        return df

    @staticmethod
    def _avro(filepath_or_buffer, nrows=None, *args, **kwargs):
        kwargs.pop("n_partitions", None)

        if is_url(filepath_or_buffer):
            s = requests.get(filepath_or_buffer).text
            df = pdx.read_avro(StringIO(s), *args, **kwargs)
        else:
            df = pdx.read_avro(filepath_or_buffer, *args, **kwargs)

        if nrows:
            logger.warn(f"'load.avro' on {EnginePretty.PANDAS.value} loads the whole dataset and then truncates it")
            df = df[:nrows]

        return df

    @staticmethod
    def _parquet(filepath_or_buffer, nrows=None, engine="pyarrow", *args, **kwargs):
        kwargs.pop("n_partitions", None)

        if is_url(filepath_or_buffer):
            s = requests.get(filepath_or_buffer).text
            df = pd.read_parquet(StringIO(s), engine=engine, *args, **kwargs)
        else:
            df = pd.read_parquet(filepath_or_buffer, engine=engine, *args, **kwargs)

        if nrows:
            logger.warn(f"'load.parquet' on {EnginePretty.PANDAS.value} loads the whole dataset and then truncates it")
            df = df[:nrows]

        return df

    @staticmethod
    def _xml(filepath_or_buffer, nrows=None, *args, **kwargs):
        kwargs.pop("n_partitions", None)
        
        if is_url(filepath_or_buffer):
            s = requests.get(filepath_or_buffer).text
            df = pd.read_xml(StringIO(s), *args, **kwargs)
        else:
            df = pd.read_xml(filepath_or_buffer, *args, **kwargs)

        if nrows:
            logger.warn(f"'load.xml' on {EnginePretty.PANDAS.value} loads the whole dataset and then truncates it")
            df = df[:nrows]

        return df

    @staticmethod
    def _excel(filepath_or_buffer, nrows=None, storage_options=None, *args, **kwargs):
        kwargs.pop("n_partitions", None)

        if is_url(filepath_or_buffer):
            s = requests.get(filepath_or_buffer).text
            filepath_or_buffer = StringIO(s)

        dfs = pd.read_excel(filepath_or_buffer, nrows=nrows, storage_options=storage_options, *args, **kwargs)
        sheet_names = list(pd.read_excel(filepath_or_buffer, None, storage_options=storage_options).keys())
        df = pd.concat(val_to_list(dfs), axis=0).reset_index(drop=True)

        return df, sheet_names    

    def orc(self, path, columns, storage_options=None, conn=None, n_partitions=1, *args, **kwargs):

        path = unquote_path(path)

        if conn is not None:
            path = conn.path(path)
            storage_options = conn.storage_options

        file, file_name = prepare_path(path, "orc")[0]

        try:
            df = pdx.read_orc(file_name, columns, storage_options=storage_options)
            df = PandasDataFrame(df, op=self.op)
            df.meta = Meta.set(df.meta, "file_name", file_name)

        except IOError as error:
            logger.print(error)
            raise

        return df

    @staticmethod
    def zip(zip_path, filename, dest=None, merge=False, storage_options=None, conn=None, n_partitions=1, *args, **kwargs):
        if dest is None:
            dest = str(uuid.uuid4()) + "/"

        zip_path = glob.glob(zip_path)

        dest = Path(dest).expanduser()

        # if csv concat all files
        # if json, multiple concat files

        for filename in zip_path:
            # print(filename)
            with zipfile.ZipFile(filename) as zf:
                zf.infolist()
                for member in zf.infolist():
                    # print(member.filename)
                    try:
                        zf.extract(member, dest)
                    except zipfile.error as e:
                        pass



from optimus.helpers.core import one_list_to_val, val_to_list
from optimus.helpers.logger import logger
from optimus.infer import is_numeric, is_nan
from optimus.helpers.constants import PYTHON_SHORT_TYPES


def compress_dict(lst, col_name):
    _result = {}
    for l in lst:
        _result.setdefault(col_name, []).append({"value": l[col_name], "count": l["count"]})
    return _result


def compress_list(lst):
    _result = {}
    for l in lst:
        for k, v in l.items():
            _result.setdefault(k, []).append(v)
    return _result


def parse_python_dtypes(value):
    """
    Get a spark data type from a string
    :param value:
    :return:
    """
    return PYTHON_SHORT_TYPES[value.lower()]


def parse_col_names_funcs_to_keys(data):
    """
    Helper function that return a formatted json with function:value inside columns. Transform from
    {'max_antiguedad_anos': 15,
    'max_m2_superficie_construida': 1800000,
    'min_antiguedad_anos': 2,
    'min_m2_superficie_construida': 20}

    to

    {'m2_superficie_construida': {'min': 20, 'max': 1800000}, 'antiguedad_anos': {'min': 2, 'max': 15}}

    :param data: json data
    :return: json
    """
    functions_array = ["range", "count_uniques", "count_na", "min", "max", "stddev", "kurtosis", "mean", "skewness",
                       "sum", "variance",
                       "approx_count_distinct", "countDistinct", "na", "zeros", "percentile", "count", "hist"]

    _result = {}
    for k, v in data[0].items():
        for f in functions_array:

            temp_func_name = f + "_"
            if k.startswith(temp_func_name):
                _col_name = k[len(temp_func_name):]
                if is_nan(v):
                    logger.print(
                        "'{FUNCTION}' function in '{COL_NAME}' column is returning 'nan'. Is that what you expected?. "
                        "Seems that '{COL_NAME}' has 'nan' values".format(
                            FUNCTION=f,
                            COL_NAME=_col_name))
                # If the value is numeric only get 5 decimals
                elif is_numeric(v):
                    v = round(v, 5)
                _result.setdefault(_col_name, {})[f] = v

                break

    return _result


def parse_data_types(df, value):
    """
    Get the data type from a string data type representation. for example 'int' from 'uint64'
    :param value:
    :param df:
    :return:
    """

    value = val_to_list(value)
    data_type = [df.constants.INTERNAL_TO_OPTIMUS.get(v, v) for v in value]
    return one_list_to_val(data_type)



import re

import numpy as np

# From a top point of view we organize Optimus separating the functions in dataframes and dask engines.
# Some functions are commons to pandas and dask.
from optimus.helpers.columns import parse_columns
from optimus.helpers.constants import Actions
from optimus.helpers.core import val_to_list
from optimus.infer import is_str


def word_tokenize(series):
    import nltk
    return nltk.word_tokenize(series)


def hist(series, bins, range):
    return np.histogram(series.to_float(), bins=bins, range=range)


def string_to_index(df, cols, output_cols=None, le=None, **kwargs):
    """_
    Maps a string column of labels to an ML column of label indices. If the input column is
    numeric, we cast it to string and index the string values.
    :param df: Dataframe to be transformed
    :param cols: Columns to be indexed.
    :param output_cols:Column where the output is going to be saved
    :param le: Label encoder library to make the process
    :return: Dataframe with indexed columns.
    """

    def _string_to_index(value):
        # Label encoder can not handle np.nan
        # value[value.isnull()] = 'NaN'
        return le.fit_transform(value.astype(str))

    return df.cols.apply(cols, _string_to_index, output_cols=output_cols,
                         meta_action=Actions.STRING_TO_INDEX.value,
                         mode="vectorized")


def index_to_string(df, cols, output_cols=None, le=None, **kwargs):
    """
    Maps a column of indices back to a new column of corresponding string values. The index-string mapping is
    either from the ML attributes of the input column, or from user-supplied labels (which take precedence over
    ML attributes).
    :param df: Dataframe to be transformed.
    :param input_cols: Columns to be indexed.
    :param output_cols: Column where the output is going to be saved.
    :param le:
    :return: Dataframe with indexed columns.
    """

    def _index_to_string(value):
        return le.inverse_transform(value.astype("int"))

    return df.cols.apply(cols, _index_to_string, output_cols=output_cols,
                         meta_action=Actions.INDEX_TO_STRING.value,
                         mode="vectorized")


def find(df, columns, sub, ignore_case=False):
    """
    Find the start and end position for a char or substring
    :param df: Dataframe to be transformed.
    :param columns:
    :param ignore_case:
    :param sub:
    :return:
    """

    columns = parse_columns(df, columns)
    sub = val_to_list(sub)

    def get_match_positions(_value, _separator):
        result = None
        if is_str(_value):
            regex_str = '|'.join([re.escape(s) for s in _separator])
            if ignore_case:
                regex_str = f"(?i)({regex_str})"
            regex = re.compile(regex_str)

            length = [[match.start(), match.end()] for match in
                      regex.finditer(_value)]
            result = length if len(length) > 0 else None
        return result

    dfd = df.data

    for col_name in columns:
        # Categorical columns can not handle a list inside a list as return for example [[1,2],[6,7]].
        # That could happened if we try to split a categorical column
        # dfd[col_name] = dfd[col_name].astype("object")
        dfd[col_name + "__match_positions__"] = dfd[col_name].astype("object").apply(get_match_positions,
                                                                                     args=(sub,))

    return df.new(dfd)



import ntpath

import cudf

from optimus.engines.base.io.load import BaseLoad
from optimus.engines.base.meta import Meta
from optimus.engines.cudf.dataframe import CUDFDataFrame
from optimus.helpers.functions import prepare_path, unquote_path
from optimus.helpers.logger import logger


class Load(BaseLoad):

    def json(self, path, multiline=False, n_rows=-1, *args, **kwargs):

        path = unquote_path(path)

        local_file_names = prepare_path(path, "json")
        try:
            df_list = []

            for file_name, j in local_file_names:
                df = cudf.read_json(file_name, lines=multiline, nrows=n_rows, *args, **kwargs)
                df_list.append(df)

            df = cudf.concat(df_list, axis=0, ignore_index=True)
            df = CUDFDataFrame(df, op=self.op)
            df.meta = Meta.set(df.meta, "file_name", local_file_names[0])

        except IOError as error:
            logger.print(error)
            raise
        return df

    def csv(self, filepath_or_buffer, sep=',', header=True, infer_schema=True, encoding="utf-8", null_value="None",
            n_rows=-1, cache=False,
            quoting=0, lineterminator=None, on_bad_lines='warn', keep_default_na=False, na_filter=True, data_type=None,
            *args, **kwargs):

        filepath_or_buffer = unquote_path(filepath_or_buffer)

        # file, file_name = prepare_path(path, "csv")[0]

        try:
            # TODO:  lineterminator=lineterminator seems to be broken
            if header is True:
                header = 0
            elif header is False:
                header = None
            # The str to ["str] is due to a bug in cudf https://github.com/rapidsai/cudf/issues/6606
            if data_type == str or data_type is None:
                data_type = ["str"]

            cdf = cudf.read_csv(filepath_or_buffer, sep=sep, header=header, encoding=encoding,
                                quoting=quoting, on_bad_lines=on_bad_lines,
                                keep_default_na=keep_default_na, na_values=null_value, nrows=n_rows,
                                na_filter=na_filter, dtype=data_type, *args, **kwargs)
            df = CUDFDataFrame(cdf, op=self.op)
            df.meta = Meta.set(df.meta, None,
                               {"file_name": filepath_or_buffer, "max_cell_length": df.cols.len("*").cols.max()})


        except IOError as error:
            logger.print(error)
            raise
        return df

    def _parquet(self, path, columns=None, *args, **kwargs):

        path = unquote_path(path)

        try:
            df = cudf.read_parquet(path, columns=columns, engine='pyarrow', *args, **kwargs)
            df = CUDFDataFrame(df, op=self.op)
            df.meta = Meta.set(df.meta, "file_name", path)

        except IOError as error:
            logger.print(error)
            raise

        return df

    def avro(self, path, storage_options=None, conn=None, *args, **kwargs):

        path = unquote_path(path)

        if conn is not None:
            path = conn.path(path)
            storage_options = conn.storage_options

        file, file_name = prepare_path(path, "avro")[0]

        try:
            df = cudf.read_avro(path, storage_options=storage_options, *args, **kwargs)
            df = CUDFDataFrame(df, op=self.op)
            df.meta = Meta.set(df.meta, "file_name", file_name)

        except IOError as error:
            logger.print(error)
            raise

        return df

    def orc(self, path, columns=None, storage_options=None, conn=None, *args, **kwargs):

        path = unquote_path(path)

        if conn is not None:
            path = conn.path(path)
            storage_options = conn.storage_options

        file, file_name = prepare_path(path, "orc")[0]

        try:
            df = cudf.read_orc(path, columns, storage_options=storage_options, *args, **kwargs)
            df = CUDFDataFrame(df, op=self.op)
            df.meta = Meta.set(df.meta, "file_name", file_name)

        except IOError as error:
            logger.print(error)
            raise

        return df

    @staticmethod
    def excel(path, sheet_name=0, storage_options=None, conn=None, *args, **kwargs):

        path = unquote_path(path)

        if conn is not None:
            path = conn.path(path)
            storage_options = conn.storage_options

        file, file_name = prepare_path(path, "xls")

        try:
            pdf = cudf.read_excel(file, sheet_name=sheet_name, storage_options=storage_options, *args, **kwargs)

            # exception when Spark try to infer the column data type
            # Parse object column data type to string to ensure that Spark can handle it. With this we try to reduce
            col_names = list(pdf.select_dtypes(include=['object']))

            column_dtype = {}
            for col in col_names:
                column_dtype[col] = str

            # Convert object columns to string
            pdf = pdf.astype(column_dtype)

            df = cudf.from_pandas(pdf, npartitions=3)
            df.meta = Meta.set(df.meta, "file_name", ntpath.basename(file_name))
        except IOError as error:
            logger.print(error)
            raise

        return df



import os

import pandavro as pdx

from optimus.engines.base.io.save import BaseSave
from optimus.helpers.logger import logger

from optimus.helpers.types import *
from optimus.engines.base.io.save import BaseSave


class Save(BaseSave):
    def __init__(self, root: 'DataFrameType'):
        self.root = root

    def json(self, path, orient="records", *args, **kwargs):

        df = self.root.data

        try:
            df.to_json(path, orient=orient, *args, **kwargs)

        except IOError as e:
            logger.print(e)
            raise

    def csv(self, path, mode="w", show_path=False, encoding="utf-8-sig", **kwargs):

        try:
            df = self.root.data
            # Dask reference
            # https://docs.dask.org/en/latest/dataframe-api.html#dask.dataframe.to_csv
            # df.to_csv(filename=path)

            if show_path is True:
                print(os.path.abspath(path))

            df.to_csv(path, index=False, mode=mode, encoding=encoding, **kwargs)

        except IOError as error:
            logger.print(error)
            raise

    def excel(self, path, mode="w", *args, **kwargs):

        try:
            df = self.root.data
            df.to_excel(path, index=False)

        except IOError as error:
            logger.print(error)
            raise

    def parquet(self, path, *args, **kwargs):
        """
        Save data frame to a parquet file
        :param path: path where the spark will be saved.
        :param mode: Specifies the behavior of the save operation when data already exists.
                    "append": Append contents of this DataFrame to existing data.
                    "overwrite" (default case): Overwrite existing data.
                    "ignore": Silently ignore this operation if data already exists.
                    "error": Throw an exception if data already exists.
        :param num_partitions: the number of partitions of the DataFrame
        :return:
        """

        # This character are invalid as column names by parquet
        invalid_character = [" ", ",", ";", "{", "}", "(", ")", "\n", "\t", "="]

        def func(col_name):
            for i in invalid_character:
                col_name = col_name.replace(i, "_")
            return col_name

        df = self.root.cols.rename(func=func)

        try:
            df.data.to_parquet(path)
        except IOError as e:
            logger.print(e)
            raise

    def avro(self, path, *args, **kwargs):
        pdx.to_avro(path, self.root.data)



from __future__ import annotations
import optimus.engines.base.engine as engine_module
import optimus.engines.base.basedataframe as dataframe_module
from typing import TypeVar, List

EngineType = TypeVar("EngineType", bound='engine_module.BaseEngine')
InternalEngineType = TypeVar("InternalEngineType")

DataFrameType = TypeVar("DataFrameType", bound='dataframe_module.BaseDataFrame')
InternalDataFrameType = TypeVar("InternalDataFrameType", bound=dict)
MaskDataFrameType = TypeVar("MaskDataFrameType", bound='dataframe_module.BaseDataFrame')
ConnectionType = TypeVar("ConnectionType")
ClustersType = TypeVar("ClustersType")
ModelType = TypeVar("ModelType")

DataFrameTypeList = TypeVar("DataFrameTypeList", bound=List['dataframe_module.BaseDataFrame'])
InternalDataFrameTypeList = TypeVar("InternalDataFrameTypeList", bound=list)
MaskDataFrameTypeList = TypeVar("MaskDataFrameTypeList", bound=list)
ConnectionTypeList = TypeVar("ConnectionTypeList", bound=list)
ClustersTypeList = TypeVar("ClustersTypeList", bound=list)
ModelTypeList = TypeVar("ModelTypeList", bound=list)

StringsList = TypeVar("StringsList", List[str], str)
StringsListNone = TypeVar("StringsListNone", List[str], str, None)

_list_types = [str(DataFrameTypeList), str(InternalDataFrameTypeList), str(MaskDataFrameTypeList), str(ConnectionTypeList), str(ClustersTypeList), str(ModelTypeList)]

_types = [str(DataFrameType), str(InternalDataFrameType), str(MaskDataFrameType), str(ConnectionType), str(ClustersType), str(ModelType)]


def is_list_of_optimus_type(value):
    return str(value) in _list_types


def is_optimus_type(value):
    return str(value) in _types


def is_any_optimus_type(value):
    return str(value) in [*_types, *_list_types]



from optimus.tests.base import TestBase


class TestRenamePandas(TestBase):
    dict = {"foobar": [1], "zoobat": [1], "foobat": [1], "john": [1]}

    def test_names_auto_regex(self):

        df = self.df

        result = df.cols.names(".*bat")

        expected = ["zoobat", "foobat"]

        self.assertEquals(result, expected)

    def test_rename_regex(self):

        df = self.df

        result = df.cols.rename(".*bat", func=lambda a: a+"bat").cols.names()

        expected = ["foobar", "zoobatbat", "foobatbat", "john"]

        self.assertEquals(result, expected)

    def test_rename(self):

        df = self.df

        result_single = df.cols.rename("foobar", "FB").cols.names()
        result = df.cols.rename(["foobar", "zoobat"], ["FB", "ZB"]).cols.names()
        result_2 = df.cols.rename([("foobar", "FB"), ("zoobat", "ZB")]).cols.names()
        result_3 = df.cols.rename({"foobar": "FB", "zoobat": "ZB"}).cols.names()

        expected_single = ["FB", "zoobat", "foobat", "john"]
        expected = ["FB", "ZB", "foobat", "john"]

        self.assertEquals(result_single, expected_single)
        self.assertEquals(result, expected)
        self.assertEquals(result_2, expected)
        self.assertEquals(result_3, expected)


class TestRenameDask(TestRenamePandas):
    config = {'engine': 'dask', 'n_partitions': 1}


class TestRenamePartitionDask(TestRenamePandas):
    config = {'engine': 'dask', 'n_partitions': 2}


try:
    import cudf # pyright: reportMissingImports=false
except:
    pass
else:
    class TestRenameCUDF(TestRenamePandas):
        config = {'engine': 'cudf'}


try:
    import dask_cudf # pyright: reportMissingImports=false
except:
    pass
else:
    class TestRenameDC(TestRenamePandas):
        config = {'engine': 'dask_cudf', 'n_partitions': 1}


try:
    import dask_cudf # pyright: reportMissingImports=false
except:
    pass
else:
    class TestRenamePartitionDC(TestRenamePandas):
        config = {'engine': 'dask_cudf', 'n_partitions': 2}

try:
    import pyspark
except:
    pass
else:
    class TestRenameSpark(TestRenamePandas):
        config = {'engine': 'spark'}

try:
    import vaex
except:
    pass
else:
    class TestRenameVaex(TestRenamePandas):
        config = {'engine': 'vaex'}